//-------------------- Copyright (C) 2009 by Geoffrey Prewett -----------------
#include "AnnotatedOutput.h"
#include "Namespace.h"
#include "Class.h"
#include "Member.h"
#include "Function.h"
#include "Global.h"

#include <sstream>

using namespace std;

AnnotatedOutput::AnnotatedOutput(ofstream& stream, Namespace& ns,
				 const char *outputFile)
    : _namespace(ns), _filename(outputFile), _out(stream), _needsLineNo(true),
      _lineNo(0)
{
}

AnnotatedOutput::~AnnotatedOutput()
{
}

void AnnotatedOutput::start()
{
    printLineNo(++_lineNo);
}

void AnnotatedOutput::keywordToken(const std::string& t)
{
    _out << "<span class=\"keyword\">" << t << "</span>";
}

void AnnotatedOutput::builtinTypeToken(const std::string& t)
{
    _out << "<span class=\"keywordtype\">" << t << "</span>";
}

void AnnotatedOutput::objectToken(const std::string& t, Symbol *symbol)
{
    if (t == "this") {
	keywordToken(t);
	return;
    }
    if (t == "super" && symbol && symbol->symbolType() == Symbol::kClass) {
	string tag = t;
	tag += "<span>&nbsp;";
	tag += symbol->name();
	tag += "&nbsp;</span>";
	_out << ::getClassLink(*(Class *)symbol, tag);
	return;
    }

    if (symbol) {
	if (symbol->symbolType() == Symbol::kClass) {
	    _out << "<span>" << html(t) << "<span>&nbsp;" << symbol->name()
		 << "&nbsp;</span></span>";
	}
	else if (symbol->symbolType() == Symbol::kMember) {
	    string tag;
	    tag += "<span>&nbsp;";
	    tag += html(((Member*)symbol)->type());
	    tag += "&nbsp;</span>";
	    tag += html(t);
	    _out << ::getMemberLink(_namespace, (Member*)symbol,
				    (symbol->definitionFile() != _filename),
				    tag, "var");
	}
	else // In case this was mis-recorded as a function or something...
//	    _out << "<span style=\"color: #880033;\">" << html(t) << "</span>";
	    _out << html(t);
    }
    else
//	_out << "<span style=\"color: #880033;\">" << html(t) << "</span>";
	_out << html(t);
}

void AnnotatedOutput::classToken(const Class *c, const std::string& t)
{
    _out << ::getClassLink(*c, html(t));
}

void AnnotatedOutput::stringToken(const std::string& t)
{
    std::string str = html(t);
    std::string::size_type pos = 0;
    while ((pos = str.find(' ')) != std::string::npos)
	str.replace(pos, 1, "&nbsp;");
    _out << "<span class=\"string\">" << str << "</span>";
}

void AnnotatedOutput::whitespaceToken(const std::string& t)
{
    unsigned int len = t.size();
    for (unsigned int i = 0;  i < len;  ++i) {
	if (t[i] == '\n') {
	    _out << "<br>" << endl;
	    printLineNo(++_lineNo);
	}
	else if (t[i] == '\t') {
	    for (int j = 0;  j < gSettings.tabSize;  ++j)
		_out << "&nbsp;";
	} else
	    _out << "&nbsp;";
    }
}

void AnnotatedOutput::commentToken(const std::string& t)
{
    _out << "<span class=\"comment\">";
    for (std::string::const_iterator it = t.begin();  it != t.end();  ++it) {
	if (*it == '\n') {
	    _out << "<br>";
	    printLineNo(++_lineNo);
	}
	else if (*it == ' ' || *it == '\t')
	    _out << "&nbsp;";
	else
	    _out << *it;
    }
    _out << "</span>";
}

void AnnotatedOutput::otherToken(const std::string& t)
{
    _out << html(t);
}

void AnnotatedOutput::operatorToken(const std::string& t)
{
    
    _out << html(t);
}

void AnnotatedOutput::functionDefinition(const std::string& className,
					 const std::string& functionName)
{
    Name actualClass;
    Member *m;
    if (_namespace.lookup(className, functionName, Symbol::kFunction,
			  &actualClass, &m)) {
	_out << "<a name=\"" << m->anchor() << "\"></a>";
//	_out << "[" << m->anchor() << "] ";
    }
    _out << functionName;
}

void AnnotatedOutput::functionCall(const std::string& caller,
                                   const std::string& callee, bool callThru,
				   const std::string& t)
{
    string::size_type funcBegin = callee.rfind('.');
    string calleeClass = callee.substr(0, funcBegin);
    string calleeName = callee.substr(funcBegin + 1);
    Name actualClass;
    Member *m = NULL;
    if (_namespace.lookup(calleeClass, calleeName, Symbol::kFunction,
			  &actualClass, &m) && m->seenDefinition()) {
	stringstream tag;
	tag << t << "<span><table class=\"func-tip\"><tr><td>&nbsp;"
	    << html(_namespace.get(m->type()).shortName())
	    << "&nbsp;<b class=\"unbold\">"
	    << html(_namespace.get(actualClass).shortName()) << ".</b>"
	    << calleeName << "(</td><td>";
	// Member variables and functions can share the same name, but
	// we don't handle it, so you might get something that isn't a
	// function.
	if (m->symbolType() != Symbol::kFunction) {
	    _out << t;
	    return;
	}
	Function::Args args = ((Function *)m)->args();
	Function::Args::iterator arg = args.begin();
	for (;  arg != args.end();  ++arg) {
	    if (arg != args.begin())
		tag << ", ";
	    if (arg->flags & Function::Arg::kConst)
		tag << "<b class=\"unbold\">const</b>&nbsp;";
	    if (arg->flags & Function::Arg::kFinal)
		tag << "<b class=\"unbold\">final</b>&nbsp;";
	    tag << html(arg->type) << "&nbsp;" << arg->name;
	}
	tag << ")";
	if (m->modifiers() & Member::kConst)
	    tag << "&nbsp;const";
	tag << "</td></tr></table></span>";

	_out << ::getMemberLink(_namespace, m,
				(m->definitionFile() != _filename),
				tag.str(), "func");
    }
    else
//	_out << "<span style=\"color: #660000;\">" << t
//	     << "<span style=\"color: black;\">caller: " << caller
//	     << "<br>callee: " << callee << "<br>class: " << calleeClass
//	     << ", name: " << calleeName << ", m: " << m << "</span></span>";
	_out << t;
}

void AnnotatedOutput::objectDefinition(const std::string& className,
				       const std::string& objectName)
{
    Name actualClass;
    Member *m;
    if (_namespace.lookup(className, objectName, Symbol::kMember,
			  &actualClass, &m)) {
	_out << "<a name=\"" << m->anchor() << "\"></a>";
//	_out << "[" << m->anchor() << "] ";
    }
    _out << objectName;
}

void AnnotatedOutput::printLineNo(int n)
{
    int max = 10000;
    while (n < max) {
	_out << "&nbsp;";
	max /= 10;
    }
    _out << n << "&nbsp;&nbsp;&nbsp;";
}
