//-------------------- Copyright (C) 2009 by Geoffrey Prewett -----------------
#include "Class.h"
#include "Namespace.h"
#include "Global.h"
#include "Exception.h"
#include "ParserFactory.h"
#include "Parser.h"
#include "AnnotatedOutput.h"

#include <stdlib.h>

#include <map>
#include <iostream>

using namespace std;

Class::~Class()
{
}

Name Class::concreteParent(Namespace& ns)
{
    Name p;
    for (unsigned int i = 0;  i < _parents.size();  ++i) {
	if (!ns.get(_parents[i]).isInterface()) {
	    if (p.empty())
		p = _parents[i];
	    else
		return "";  // Multiple concrete parents
	}
    }
    return p;
}

void Class::getAllMembers(Namespace& ns, std::vector<const Member*> *members)
{
    members->clear();

    if (!_allMembersCache.empty()) {
	*members = _allMembersCache;
	return;
    }


    // Get all the members from the ancestors.  We are not going to
    // worry about what happens if two interfaces define the same function;
    // we'll only pick one of them.  That would likely cause enough problems
    // that the programmers will fix their code, so it shouldn't worry us much.
    vector<Name>::iterator it;
    for (it = _parents.begin();  it != _parents.end();  ++it)
	ns.get(*it).getAllMembers(ns, members);

    // Put everything in a set so that we can properly override members
    map<Name, const Member*> m;
    for (unsigned int i = 0;  i < members->size();  ++i)
	m[(*members)[i]->mangledName()] = (*members)[i];

    // Add and override this classes members;
    Members::const_iterator mit;
    for (mit = _members.begin();  mit != _members.end();  ++mit) {
	m[mit->second->mangledName()] = mit->second;
    }

    
    map<Name, const Member*>::iterator copyIt;
    members->clear();
    for (copyIt = m.begin();  copyIt != m.end();  ++copyIt)
	members->push_back(copyIt->second);

    // Cache if this isn't a leaf class
    if (!_children.empty())
	_allMembersCache = *members;
}

bool Class::writeHierarchyGraph(Namespace& ns, const string& outputDir)
{
    ofstream dotInput((outputDir + this->hierarchyGraphDotFilename()).c_str());
    if (!dotInput.is_open())
	return false;

    dotInput << "digraph hierarchy {" << endl;
    dotInput << "   graph [bgcolor=transparent,rankdir=BT];" << endl;
    dotInput << "   node [shape=box,fontsize=10,fontname=Arial,height=.25,style=filled,fillcolor=white];"
	     << endl;
    dotInput << "   \"" << _shortName << "\" [fillcolor=\"gray\"];" << endl;

    outputParentEdges(dotInput, ns);

    // Output classes that inherit from this class (but only one level deep)
    for (unsigned int i = 0;  i < _children.size();  ++i) {
	Class& child = ns.get(_children[i]);
	dotInput << "   \"" << child.shortName() << "\"->\"" << _shortName
		 << "\"" << endl;
	if (child.seenDefinition())
	    dotInput << "   \"" << child.shortName()
		     << "\" [URL=\"javascript:load('" << child.classFilename()
		     << "', '" << child.sourceFilename() << "', '')\"]" << endl;
    }

    dotInput << "}" << endl;
    dotInput.close();

    string cmd = string("dot -Tcmapx -o ") + outputDir + this->hierarchyGraphImageMapFilename() + string(" ") + outputDir + this->hierarchyGraphDotFilename();
    int retVal = system(cmd.c_str());
    if (retVal != 0)
	return false;
    return (system((string("dot -Tpng -o ") + outputDir + this->hierarchyGraphFilename() + string(" ") + outputDir + this->hierarchyGraphDotFilename()).c_str()) != 0);
}

void Class::outputParentEdges(ofstream& dotInput, Namespace& ns)
			      
{
    if (_parents.empty())
	return;

    for (unsigned int i = 0;  i < _parents.size();  ++i) {
	if (_parents[i] != this->name()) {
	    Class& p = ns.get(_parents[i]);
	    Name parentName = p.shortName();
	    if (!p.seenDefinition())
		parentName = _parents[i];
	    if (p.seenDefinition())
		dotInput << "   \"" << parentName
			 << "\" [URL=\"javascript:load('" << p.classFilename()
			 << "', '" << p.sourceFilename() << "', '')\"]" << endl;
	    dotInput << "   \"" << shortName() << "\"->\"" << parentName
		     << "\"" << endl;
	    p.outputParentEdges(dotInput, ns);
	}
    }
}
