//-------------------- Copyright (C) 2009 by Geoffrey Prewett -----------------
#ifndef CLASS_H
#define CLASS_H

#include "Global.h"
#include "Name2Type.h"
#include "Symbol.h"
#include "Member.h"
#include "NameTable.h"

#include <string>
#include <vector>
#include <fstream>

class Namespace;

class Class : public Symbol
{
 public:
    typedef std::vector<Name> NameArray;
    typedef NameTable<Member> Members;

    Class() : Symbol(), _isInterface(false)/*, _definedAtLine(-1)*/ {}
    ~Class();

    virtual SymbolType symbolType() const { return kClass; }

    /** Returns the full name of the class (including the namespace/package) */
//    const Name& name() const { return _name; }
    virtual void setName(const Name& n) { Symbol::setName(n);  _shortName = n; }

    /** Returns the name of the class (does not include the namespace/package)*/
    const Name& shortName() const { return _shortName; }
    void setShortName(const Name& n) { _shortName = n; }

    const std::vector<Name>& parents() const { return _parents; }
    void addParent(const Name& p) { _parents.push_back(p); }
    /** Returns the name of the concrete parent if there is one.  If not,
	or the parents are not known, returns empty. */
    Name concreteParent(Namespace& ns);

    /** Returns the name of the template parameter
	(e.g. class c<T> should return "T", assuming the parser did it's job) */
    const Name& templateParameter() { return _templateParam; }
    void setTemplateParameter(const Name& n) { _templateParam = n; }

    const std::vector<Name>& children() const { return _children; }
    void addChild(const Name& c) { _children.push_back(c); }

    void setInterface() { _isInterface = true; }
    bool isInterface()
    {
	if (seenDefinition())
	    return _isInterface;
	return true;  // If we haven't seen this, assume it's an interface
    }

    void setOuterClass(const Name& c) { _outerClass = c; }
    const Name& outerClass() const { return _outerClass; }
    bool isInnerClass() const { return (!_outerClass.empty()); }

    void addInnerClass(const Name& c) { _innerClasses.push_back(c); }
    const NameArray& innerClasses() { return _innerClasses; }
    

    /** Returns the members of this class */
    Members& members()
	{ _allMembersCache.clear();  return _members; }
    const Members& members_const() const { return _members; }
    
    /** Returns all members, including inherited ones */
    void getAllMembers(Namespace& ns, std::vector<const Member*> *members);

//    void setDefinition(const std::string& file, int line)
//	{ _definedInFile = file;  _definedAtLine = line; }
//    const std::string& definitionFile() const{ return _definedInFile; }
//    int definitionLine() const { return _definedAtLine; }
    
//    bool seenClassDefinition() const { return _definedAtLine > 0; }

//    std::string sourceFilename() { return _name + std::string("-src.html"); }
    std::string classFilename() const { return name() + std::string(".html"); }
    std::string hierarchyGraphFilename() const { return name() + std::string("-hierarchy.png"); }
    std::string hierarchyGraphImageMapFilename() const { return name() + std::string("-hierarchy.map"); }
    std::string hierarchyGraphDotFilename() const { return name() + std::string("-hierarchy.dot"); }

    /** Writes a hierarchy graph to outputDir + hierarchyGraphFilename()
	(Note that outputDir should include a trailing slash! */
    bool writeHierarchyGraph(Namespace& ns, const std::string& outputDir);

 private:
//    Name _name;  // Fully-qualified name of the class
    Name _shortName;  // Just the class name
    std::vector<Name> _parents;
    std::vector<Name> _children;
    Name _outerClass;
    NameArray _innerClasses;
    bool _isInterface;
    Members _members;
//    std::string _definedInFile;
//    int _definedAtLine;
    std::vector<const Member*> _allMembersCache; // To speed up deep hierarchies
    Name _templateParam;

    void outputParentEdges(std::ofstream& dotInput, Namespace& ns);
};

#endif // CLASS_H
