//-------------------- Copyright (C) 2009 by Geoffrey Prewett -----------------
#include "Graph.h"

#include <iostream>
#include <sstream>

using namespace std;

void Graph::Node::print() const {
    std::map<Name, Counter>::const_iterator it;
    cerr << "Node " << name << endl << "   In edges:" << endl;
    for (it = in.begin();  it != in.end();  ++it)
	cerr << "      " << it->first << endl;
    cerr << "   Out edges:" << endl;
    for (it = out.begin();  it != out.end();  ++it)
	cerr << "      " << it->first << endl;
}

void Graph::add(const std::string& caller, const std::string& callee,
		const std::string& mods)
{
    std::string callerObj, callerMethod, calleeObj, calleeMethod;
    getObject(caller, &callerObj, &callerMethod);
    getObject(callee, &calleeObj, &calleeMethod);
    if (gSettings.debug)
	cerr << "   (debug) Graph::add(" << caller << " -> " << callee
	     << ")" << endl;

    if (!_namespace.get(calleeObj).seenDefinition() &&
	!gSettings.showUndefinedClasses) {
	if (gSettings.debug)
	    cerr << "  rejected:  haven't seen class definition yet" <<endl;
	return;
    }
    if (calleeObj == caller) {  // Constructor calling into same class
	if (gSettings.debug)
	    cerr << "  rejected:  constructor calling class function"<<endl;
	return;
    }

    // If they aren't the same object, and caller isn't calling super,
    // add the call
    if (callerObj != calleeObj &&
	!_namespace.instanceOf(callerObj, calleeObj)) {
	stringstream s;
	s << "\"" << caller << "\"->\"" << callee << "\" " << mods; 
	std::map<std::string, int>::iterator it;
	it = _externalCalls.find(s.str());
	if (it != _externalCalls.end())
	    _externalCalls[s.str()]++;
	else
	    _externalCalls[s.str()] = 1;
	_usedMethods.insert(callerMethod);
	_usedMethods.insert(calleeMethod);

	if (_nodes.find(callerObj) == _nodes.end())
	    _nodes[callerObj] = Node(callerObj);
	if (_nodes.find(calleeObj) == _nodes.end())
	    _nodes[calleeObj] = Node(calleeObj);
	_nodes[callerObj].out[calleeObj]++;
	_nodes[calleeObj].in[callerObj]++;
    }
}

void Graph::getObject(const std::string& call,
		      std::string *obj, std::string *method)
{
    string::size_type pos = call.rfind(".");
    if (pos == string::npos) { // Must be a constructor
	*obj = call;
	*method = call;
    } else {
	*obj = call.substr(0, pos);
	*method = call.substr(pos + 1);
    }
}

void Graph::reduceToCycles()
{
    int nRemoved;
    std::map<Name, Node>::iterator it;
    std::map<Name, Counter>::iterator it2;

    do {
	nRemoved = 0;
	for (it = _nodes.begin();  it != _nodes.end();  ++it) {
	    if (it->second.edges() > 0 && it->second.onlyInOrOut()) {
		for (it2 = it->second.in.begin();
		     it2 != it->second.in.end();  ++it2)
		    _nodes[it2->first].out.erase(it->first);
		for (it2 = it->second.out.begin();
		     it2 != it->second.out.end();  ++it2)
		    _nodes[it2->first].in.erase(it->first);
		it->second.in.clear();
		it->second.out.clear();
		nRemoved++;
	    }
	}
    } while(nRemoved);
}

void Graph::print()
{
    std::map<std::string, int>::iterator it;
	
    cout << "digraph G {" << endl;
    for (it = _externalCalls.begin();  it != _externalCalls.end();  ++it) {
	if (it->second > 1) {
	    unsigned int size = it->first.size();
	    if (it->first[size - 1] == ']')
		cout << "   " << it->first.substr(0, size - 1)
		     << ",label=\"" << it->second << "\"];" << endl;
	    else
		cout << "   " << it->first << " [label=\"" << it->second
		     << "\"];" << endl;
	} else
	    cout << "   " << it->first << ";" << endl;
    }

    _namespace.printSubGraphs();

    cout << "}" << endl;
}

void Graph::printClasses()
{
    cout << "digraph G {" << endl;
    cout << "   graph [label=\"" << gSettings.title
	 << "\",fontsize=16,fontname=Arial,labelloc=top];" << endl;
    cout << "   node [fontsize=8,fontname=Arial];" << endl;
    cout << "   edge [fontsize=8,fontname=Arial];" << endl;

    std::map<std::string, Counter> classExternal;

    for (std::map<std::string, int>::iterator it = _externalCalls.begin();
	 it != _externalCalls.end();  ++it) {
	string::size_type pos = it->first.rfind("->");
	string caller = it->first.substr(1, (pos - 1) - 1);
	string callee = it->first.substr(pos + 3, it->first.rfind("\"") - (pos + 3));
	std::string callerObj, callerMethod, calleeObj, calleeMethod;
	getObject(caller, &callerObj, &callerMethod);
	getObject(callee, &calleeObj, &calleeMethod);
	    
	stringstream s;
	s << "\"" << callerObj << "\"->\"" << calleeObj << "\"";
	classExternal[s.str()]++;
    }

    if (gSettings.onlyShowCycles)
	reduceToCycles();

    std::set<Name> classUsed;
    for (std::map<std::string, Counter>::iterator it= classExternal.begin();
	 it != classExternal.end();  ++it) {
	string::size_type pos = it->first.rfind("->");
	string caller = it->first.substr(1, (pos - 1) - 1);
	string callee = it->first.substr(pos + 3, it->first.rfind("\"") - (pos + 3));
	// Two classes that have just one call between them and no
	// calls to anything else are not too interesting and just
	// take up space...
	if (it->second == 1 &&
	    _nodes[caller].edges() == 1 && _nodes[callee].edges() == 1)
	    continue;
	if (_nodes[caller].edges() < gSettings.minEdges ||
	    _nodes[callee].edges() < gSettings.minEdges)
	    continue;
	// This is kind of redudant;  reduceToCycles() would do the
	// same thing if only iterated one time
	if (gSettings.noEndOrStartNodes &&
	    (_nodes[caller].onlyInOrOut() || _nodes[callee].onlyInOrOut()))
	    continue;
	classUsed.insert(caller);
	classUsed.insert(callee);

	cout << "   " << it->first;
	int weight = 1;
	string color = "black";
	if (it->second >= 50) {
	    weight = 16;
	    color = "red2";
	} else if (it->second >= 20) {
	    weight = 8;
	    color = "red2";
	} else if (it->second >= 15) {
	    weight = 6;
	    color = "darkorange";
	} else if (it->second >= 10) {
	    weight = 4;
	    color = "springgreen4";
	} else if (it->second >= 5) {
	    weight = 2;
	    color = "mediumblue";
	} else if (it->second >= 2) {
	    weight = 1;
	    color = "darkorchid4";
	}
	if ((_nodes[caller].edges() > 10 || _nodes[callee].edges() > 10) &&
	    weight < 5)
	    weight = 5;
	if (it->second > 1)
	    cout << "[label=\"" << it->second << "\",weight=" << weight
		 << ",color=\"" << color << "\"]";
	cout << ";" << endl;
    }

    if (gSettings.showHierarchy)
	_namespace.printHierarchy(&classUsed);

    // Make classes that have lots of edges have bigger nodes
    // and undefined classes colored grey
    std::map<std::string, Node>::iterator it2;
    for (it2 = _nodes.begin();  it2 != _nodes.end();  ++it2) {
	if (classUsed.find(it2->first) == classUsed.end())
	    continue;
	cout << "   \"" << it2->first << "\" [label=\""
	     << makePrettyName(it2->first) << "\"";
	if (!_namespace.get(it2->first).seenDefinition())
	    cout << ",color=\"gray\"";
	if (it2->second.edges() > 10)
	    cout << ",height=1";
	cout << "];" << endl;
    }

    cout << "}" << endl;
}
