//-------------------- Copyright (C) 2009 by Geoffrey Prewett -----------------
#include <stdio.h> // For file operations;  printf, etc. not used
#include <errno.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <assert.h>

#include <iostream>
#include <string>
#include <map>
#include <set>
#include <vector>
#include <sstream>
#include <fstream>
#include <algorithm>

#include "Global.h"
#include "Exception.h"
#include "Counter.h"
#include "Namespace.h"
#include "ScopedNames.h"
#include "AnnotatedOutput.h"
#include "Graph.h"
#include "JavaParser.h"
#include "ParserFactory.h"
#include "CallTreeOutput.h"

using namespace std;

static const char* kStyleSheetFile = "styles.css";
static const char* kJavascriptFile = "javascript.js";
static const char* kClassIndexFile = "classes.html";
static const char* kFileIndexFile = "files.html";
static const char* kIndexHTML = "index.html";
static const unsigned int kMaxCallersInGraph = 20;
static const int kNColumnsInCallerListing = 2;

class JavaBrowser
{
 public:
    JavaBrowser(): _callers(_ns) {}

    void setOutputDir(const string& dir) { _outputDir = dir + "/"; }
    void addExtension(const char *extension)
	{ _validExtensions.push_back(extension); }
    void addFiles(const string &entity, const string& dir);

    bool parseFiles();
    bool writeBrowser();
    bool writeClassIndex();
    bool writeFileIndex();
    bool writeClassFile(Class& c);
    bool writeSourceFile(Class& c);
    bool writeCallGraph(Function *f);

 private:
    struct MemberSortByName {
	const Member *m;
	MemberSortByName() : m(NULL) {}
	MemberSortByName(const Member *member) : m(member) {}
	bool operator<(const MemberSortByName& that) const
	    { return this->m->name() < that.m->name(); }
    };
    struct ClassSortByShortName {
	Class *c;
	ClassSortByShortName() : c(NULL) {}
	ClassSortByShortName(Class *c_) : c(c_) {}
	bool operator<(const ClassSortByShortName& that) const
	    { return this->c->shortName() < that.c->shortName(); }
    };

    vector<string> _validExtensions;
    vector<string> _files;
    string _outputDir;
    Namespace _ns;
    CallTreeOutput _callers;

    bool isParseable(const string& name) const;

    bool prepareNamespace(vector<string>& files);
    bool includeFile(ofstream& out, const string& file);
    string getLink(const Name& type);
    void writeMembers(ofstream& out, vector<MemberSortByName>& members,
		      const Name& classname);
    void writeHeader(ofstream& out, const std::string& title);
    void writeStyleSheet();
    void writeJavascript();
};

void printUsage(char *progName)
{
    cout << "Usage:  " << progName << " [options] dir1 [dir2 ...]" << endl;
    cout << "   Creates HTML code browsing for .java files in the directories"
	 << endl
         << "   provided in dirs.  The output-dir is \".\" by default." << endl;
    cout << "   --debug             Enables debugging output" << endl;
    cout << "   --title Title       Sets the web site title to \"Title\"." << endl;
    cout << "                       Use double quotes if Title has spaces" << endl;
    cout << "   --output-dir dir    Write html files in dir/  (Default: \".\")" << endl;
    cout << "   --tab-size n        Sets the tab length to be n.  (Default: 8)" << endl;
}

int main(int argc, char *argv[])
{
    if (argc < 2) {
        printUsage(argv[0]);
        return 0;
    }

#ifdef WIN32
    char dotTestCmd[] = "dot -V >NUL";
#else
    char dotTestCmd[] = "dot -V 2> /dev/null";
#endif
    if (system(dotTestCmd) != 0) {
	cerr << "`dot` is not in the path!  Please install the graphviz package." << endl;
	return 1;
    }

    gSettings.debug = false;
    gSettings.minEdges = 1;
    gSettings.noEndOrStartNodes = false;
    gSettings.onlyShowCycles = false;
    gSettings.showHierarchy = false;
    gSettings.showUndefinedClasses = false;
    gSettings.title = "Source code";
    gSettings.tabSize = 8;

    int start = 1;
    string outputDir = ".";
    while (start < argc && argv[start][0] == '-' && argv[start][1] == '-') {
        if (string(argv[start]) == "--debug")
            gSettings.debug = true;
	else if (string(argv[start]) == "--output-dir") {
	    outputDir = argv[++start];
	    if (mkdir(outputDir.c_str(), S_IRWXU | S_IRWXG | S_IRWXO) != 0) {
		if (errno == EEXIST)
		    cerr << "Warning:  output dir (" << outputDir
			 << ") already exists" << endl;
		else {
		    cerr << "Could not create output directory \""
			 << outputDir << "\"!" << endl;
		    return 1;
		}
	    }
	} else if (string(argv[start]) == "--title")
	    gSettings.title = argv[++start];
	else if (string(argv[start]) == "--tab-size")
	    gSettings.tabSize = atoi(argv[++start]);
	else {
	    printUsage(argv[0]);
	    return 0;
	}
	start++;
    }
    gSettings.title += " (Generated by CodeFalcon)";

    JavaBrowser browser;

    browser.addExtension(".java");
    browser.setOutputDir(outputDir);
    for (int i = start;  i < argc;  ++i)
	browser.addFiles(argv[i], "");

    browser.parseFiles();
    browser.writeBrowser();
}

void JavaBrowser::addFiles(const string& entity, const string& dir)
{
    struct stat info;
    string filename = dir + entity;
    if (stat(filename.c_str(), &info)) {
	cerr << "Could not open '" << filename 
	     << "' (stat() failed with errno " << errno << ")" << endl;
	return;
    }
    if (S_ISDIR(info.st_mode)) {
	DIR *d = opendir(filename.c_str());
	struct dirent *dinfo;
	string dirname = entity;
	string::size_type pos = dirname.rfind('/');
	if (pos != string::npos)
	    dirname = dirname.substr(pos + 1);
	string newDir = dir + dirname + string("/");
	while ((dinfo = readdir(d))) {
	    if (dinfo->d_name != string(".") &&
		dinfo->d_name != string("..")) {
		addFiles(dinfo->d_name, newDir);
	    }
	}
	closedir(d);
    } else if (isParseable(entity)) {
	cout << "Adding '" << (dir + entity) << "'" << endl;
	_files.push_back(dir + entity);
    }
}

bool JavaBrowser::isParseable(const string& name) const
{
    string::size_type pos = name.rfind('.');
    string ext;
    if (pos != string::npos)
	ext = name.substr(name.rfind('.'));
    for (unsigned int i = 0;  i < _validExtensions.size();  ++i) {
	if (ext == _validExtensions[i])
	    return true;
    }
    return false;
}

bool JavaBrowser::parseFiles()
{
    // If we just start parsing, we won't know the full names of our
    // classes, so when we derive from them, we won't be able to find their
    // full name.  So parse through all the files quickly and populate the
    // namespace with all the full names ahead of time
    prepareNamespace(_files);
    if (gSettings.debug)
        _ns.print();

    // Parse the files to find all the classes
    ParseOutput dummy;
    for (unsigned int i = 0;  i < _files.size();  ++i) {
        try {
            Parser *p = ParserFactory::create(_files[i].c_str(), _ns, dummy);
            if (!p) {
                cerr << "Could not find a parser for " << _files[i] << endl;
                continue;
            }
            if (gSettings.debug)
              cerr << "--------------- " <<  _files[i]
		   << " (header) -------------" << endl;
            p->parse(Parser::kBuildHierarchy);
        } catch(Exception& e) {
            cerr << e.msg() << endl;
//      } catch(...) {
//          cerr << gParseState.file << ":" << gParseState.lineNo
//               << ":  Internal error:  unhandled exception" << endl;
        }
    }

    if (gSettings.debug)
        _ns.print();

    cout << "Generating call trees..." << endl;
    for (unsigned int i = 0;  i < _files.size();  ++i) {
        try {
            Parser *p = ParserFactory::create(_files[i].c_str(), _ns, _callers);
	    if (p) // No error handling, because errors already generated above
		p->parse(Parser::kOutputCallGraph);
            if (gSettings.debug)
              cerr << "--------------- " <<  _files[i]
		   << " (call tree) -------------" << endl;
	} catch(Exception& e) {
            cerr << e.msg() << endl;
	}
    }

    return true;
}

bool JavaBrowser::prepareNamespace(vector<string>& files)
{
    for (unsigned int i = 0;  i < files.size();  ++i) {
	ifstream in(files[i].c_str());
	if (!in.is_open())
	    cerr << "Could not open file " << files[i] << endl;
	string token;
	while (in.good()) {
	    in >> token;
	    if (token == "package") {
		in >> token;
		if (token[token.size() - 1] == ';')
		    token.erase(token.size() - 1);
		// Take advantage of the fact that Java files must be named
		// the same as their class.
		int start = files[i].rfind("/") + 1;
		int end = files[i].rfind(".java");
		string className = files[i].substr(start, end - start);
		className = token + string(".") + className;
		_ns.get(className);
		break;
	    }
	}
	in.close();
    }

    return true;
}

bool JavaBrowser::writeBrowser()
{
    writeStyleSheet();
    writeJavascript();

    writeClassIndex();
    writeFileIndex();

    Namespace::iterator classIt;
    for (classIt = _ns.begin();  classIt != _ns.end();  ++classIt) {
	if (classIt->seenDefinition()) {
	    std::cout << "Writing files for class " << classIt->name()
		      << endl;
	    writeClassFile(*classIt);
	    writeSourceFile(*classIt);

	    Class::Members::const_iterator mit;
	    for (mit = classIt->members_const().begin();
		 mit != classIt->members_const().end();  ++mit) {
		if (mit->second->symbolType() == Symbol::kFunction)
		    writeCallGraph((Function *)mit->second);
	    }
	}
    }

    // Write main file
    ofstream index((_outputDir + kIndexHTML).c_str());
    if (index.is_open()) {
	writeHeader(index, gSettings.title);
	index << "<frameset cols=\"40%,60%\">" << endl;
	index << "<frame src=\"" << kClassIndexFile << "\" name=\"class\">"
	      << endl;
	index << "<frameset rows=\"80%,20%\">" << endl;
	index << "<frame src=\"files.html\" name=\"src\">" << endl;
	index << "<frame src=\"\" name=\"caller\">" << endl;
	index << "</frameset>" << endl;
	index << "</frameset>" << endl;
	index << "</html>" << endl;
	
	index.close();
    }

    return true;
}

bool JavaBrowser::writeClassIndex()
{
    ofstream out((_outputDir + kClassIndexFile).c_str());
    
    if (!out.is_open())
	return false;
    cout << "Writing class index file..." << endl;

    writeHeader(out, "Class Index");
    out << "<h1><center>Class Index</center></h1>" << endl;

    vector<ClassSortByShortName> classes;
    for (Namespace::iterator it = _ns.begin();  it != _ns.end();  ++it) {
	if (it->seenDefinition() && !it->isInnerClass()) {
	    classes.push_back(&(*it));
	}
    }
    sort(classes.begin(), classes.end());

    vector<ClassSortByShortName>::iterator it;
    for (it = classes.begin();  it != classes.end();  ++it) {
	if (it->c->seenDefinition())
	    out << getClassLink(*it->c);
	out << "<br>" << endl;
    }

    out.close();
    return true;
}

bool JavaBrowser::writeClassFile(Class& c)
{
    ofstream out((_outputDir + c.name() + ".html").c_str());

    if (!out.is_open())
	return false;

    writeHeader(out, c.shortName());
    out << "<table class=\"nav\" width=\"100%\" border=\"0\" cellpadding=\"0\" cellspacing=\"0\">" << endl;
    out << "  <tr><td><center><a href=\"javascript:load('" << kClassIndexFile
	<< "','','')\">Classes</a></center></td>" << endl;
    out << "      <td><center>" << c.shortName() << "</center></td>" << endl;
    out << "      <td><center><a href=\"javascript:load('" << kFileIndexFile
	<< "','','')\">Files</a></center></td></tr>" << endl;
    out << "</table>" << endl;

    out << "<h1><center>" << c.shortName() << "</center></h1>" << endl;
    string hierarchy = c.name() + "-hierarchy";
    c.writeHierarchyGraph(_ns, _outputDir);
    out << "<center>" << endl;
    out << "<img src=\"" <<  c.hierarchyGraphFilename()
	<< "\" usemap=\"#hierarchy\" border=\"0\"><br><br>" << endl;
    out << "</center>" << endl;

    includeFile(out, _outputDir + c.hierarchyGraphImageMapFilename());

    vector<MemberSortByName> vars, publicFuncs, protFuncs, privateFuncs;
    vector<const Member*> allMembers;
    c.getAllMembers(_ns, &allMembers);
    for (vector<const Member*>::iterator it = allMembers.begin();  it != allMembers.end(); ++it) {
	const Member *m = *it;
	if (m->symbolType() == Member::kFunction) {
	    if (m->modifiers() & Member::kPublic)
		publicFuncs.push_back(MemberSortByName(m));
	    else if (m->modifiers() & Member::kProtected)
		protFuncs.push_back(MemberSortByName(m));
	    else
		privateFuncs.push_back(MemberSortByName(m));
	}
	else
	    vars.push_back(MemberSortByName(m));
    }
    sort(publicFuncs.begin(), publicFuncs.end());
    sort(protFuncs.begin(), protFuncs.end());
    sort(privateFuncs.begin(), privateFuncs.end());
    sort(vars.begin(), vars.end());

    out << "<a href=\"javascript:toggleVisibilityLink('inherited', 'showHideInherited')\" id=\"showHideInherited\">Hide inherited members</a><br><br>" << endl;

    out << "<table cellpadding=\"0\" cellspacing=\"0\">" << endl;
    if (!publicFuncs.empty()) {
	out << "   <tr><td colspan=\"2\"></td></tr>" << endl;
	out << "   <tr><td colspan=\"2\"><font size=\"5\"><b>Public Methods</b></font></td></tr>" << endl;
	writeMembers(out, publicFuncs, c.name());
    }
    if (!protFuncs.empty()) {
	out << "   <tr><td colspan=\"2\"></td></tr>" << endl;
	out << "   <tr><td colspan=\"2\"><font size=\"5\"><b>Protected Methods</b></font></td></tr>" << endl;
	writeMembers(out, protFuncs, c.name());
    }
    if (!privateFuncs.empty()) {
	out << "   <tr><td colspan=\"2\"></td></tr>" << endl;
	out << "   <tr><td colspan=\"2\"><font size=\"5\"><b>Private Methods</b></font></td></tr>" << endl;
	writeMembers(out, privateFuncs, c.name());
    }
    if (!vars.empty()) {
	out << "   <tr><td colspan=\"2\"></td></tr>" << endl;
	out << "   <tr><td colspan=\"2\"><font size=\"5\"><b>Members</b></font></td></tr>" << endl;
	writeMembers(out, vars, c.name());
    }
    out << "</table>" << endl;
    out << "<br><br>" << endl;

    if (!c.innerClasses().empty()) {
	out << "<font size=\"5\"><b>Inner Classes</b></font><br>" << endl;
	out << "<div style=\"margin-left: 40px;\">" << endl;
	const Class::NameArray& inners = c.innerClasses();
	Class::NameArray::const_iterator it;
	for (it = inners.begin();  it != inners.end();  ++it) {
	    out << getLink(*it) << "<br>" << endl;
	}
	out << "</div>" << endl;
    }

    out << "</html>" << endl;
    out.close();
}

void JavaBrowser::writeMembers(ofstream &out, vector<MemberSortByName>& members, const Name& classname)
{
    for (vector<MemberSortByName>::iterator it = members.begin();  it != members.end(); ++it) {
	const Member *m = it->m;
	bool isInherited = (m->owningClass()->name() != classname);

	out << "   <tr";
	if (isInherited)
	    out << " class=\"inherited\"";
	out << "><td class=\"memberlist\">";

	if (m->modifiers() & Member::kPureVirtual) out << "<i>";
	if (m->symbolType() & Member::kConst)
	    out << "const&nbsp;";
	out << getLink(m->type()) << "&nbsp;</td>";
	if (m->modifiers() & Member::kPureVirtual) out << "</i>";
	out << "<td>";
	if (m->modifiers() & Member::kPureVirtual) out << "<i>";
	if (m->seenDefinition())
	    out << ::getMemberLink(_ns, m, false);
	else
	    out << m->name();
	if (m->symbolType() == Member::kFunction) {
	    out << "(";
	    Function::Args args = ((Function *)m)->args();
	    Function::Args::iterator arg = args.begin();
	    for (;  arg != args.end();  ++arg) {
		if (arg != args.begin())
		    out << ",&nbsp;";
		if (arg->flags & Function::Arg::kConst)
		    out << "const&nbsp;";
		if (arg->flags & Function::Arg::kFinal)
		    out << "final&nbsp;";
		out << getLink(arg->type) << "&nbsp;" << arg->name;
	    }
	    out << ")";
	}

	string modifiers;
	if (m->modifiers() & Member::kStatic)
	    modifiers += "static&nbsp;";
	if (m->modifiers() & Member::kFinal)
	    modifiers += "final&nbsp;";
	if (m->modifiers() & Member::kPureVirtual)
	    modifiers += "abstract&nbsp;";
	if (m->symbolType() == Symbol::kMember) {
	    if (m->modifiers() & Member::kProtected)
		modifiers += "protected&nbsp;";
	    if (m->modifiers() & Member::kPrivate)
		modifiers += "private&nbsp;";
	}
	if (isInherited)
	    modifiers += getLink(m->owningClass()->name()) + " ";
	if (!modifiers.empty())
	    out << "&nbsp;&nbsp;[&nbsp;" << modifiers << "]";
	if (m->modifiers() & Member::kPureVirtual) out << "</i>";
	out << "</td></tr>" << endl;
    }
}

bool JavaBrowser::writeSourceFile(Class& c)
{
    if (!c.seenDefinition())
	return true;  // Nothing written, but no error

    std::string srcFile = _outputDir + c.sourceFilename();
    ofstream out(srcFile.c_str());
    AnnotatedOutput output(out, _ns, srcFile.c_str());
    Parser *p = ParserFactory::create(c.definitionFile().c_str(), _ns, output);
    if (!p) {
	cerr << "Could not find a parser for " << c.definitionFile() << endl;
	return false;
    }
    if (gSettings.debug)
	cerr << "--------------- " << c.definitionFile() << " (output) ---------------" << endl;

    string title = c.definitionFile().substr(c.definitionFile().rfind("/") + 1);
    writeHeader(out, title);

    out << "<table class=\"rightnav\" border=\"0\" cellpadding=\"0\" cellspacing=\"0\">" << endl;
    out << "<tr><td>&nbsp;" << title << "&nbsp;</td></tr>" << endl;
    out << "</table>" << endl;

    out << "<h1><center>" << c.definitionFile().substr(c.definitionFile().rfind("/") + 1) << "</center></h1>" << endl;
    out << "<div style=\"font-family: monospace;\">" << endl;

    try {
	p->parse(Parser::kOutputCallGraph);
    } catch(Exception& e) {
	cerr << e.msg() << endl;
	return false;
    }

    out << "</div>" << endl;
    out << "</html>" << endl;

    return true;
}

bool JavaBrowser::writeCallGraph(Function *f)
{
    // Don't write out callgraphs for functions that override a virtual method
    if (f != f->getAncestor(_ns))
	return true;

    cout << "   Writing callgraph for " << f->mangledName() << "..." << endl;

    ofstream out((_outputDir + f->callgraphFilename()).c_str());
    if (!out.is_open())
	return false;

    string fName = f->owningClass()->name() + string(".") + f->name();
    writeHeader(out, fName);

    unsigned int size = _callers.nCallers(f);
    if (size == 0) {
	out << "No callers for " << f->owningClass()->name() << "." << f->name()
	    << "</html>" << endl;
	out.close();
	return true;
    }
    if (size > kMaxCallersInGraph) {
	const CallTreeOutput::CallerSet callers = _callers.callers(f);
	CallTreeOutput::CallerSet::const_iterator it;
	out << "<h4><center>Callers for " << f->owningClass()->shortName()
	    << "." << f->name() << "</center></h4>" << endl;
	out << "<table width=\"100%\">";
	int n = 0;
	string text;
	for (it = callers.begin();  it != callers.end();  ++it) {
	    if (n % kNColumnsInCallerListing == 0) {
		if (n > 0)
		    out << "</tr>" << endl;
		out << "   <tr>";
	    }
	    n++;
	    out << "<td>";
	    if ((*it)->owningClass() != f->owningClass())
		out << (*it)->owningClass()->shortName() << ".";
	    out << getMemberLink(_ns, *it, true) << "</td>";
	}
	out << "</html>" << endl;
	out.close();
	return true;
    }

    string dotfile = _outputDir + f->callgraphDotFilename();
    string imagemap = _outputDir + f->callgraphImageMapFilename();
    string callgraph = _outputDir + f->callgraphImageFilename();
    string cmd;
    int retVal;

    ofstream dot(dotfile.c_str());
    if (!dot.is_open()) {
	goto error;
    }
    _callers.writeCallGraph(dot, f);
    dot.close();

    cmd = string("dot -Tcmapx -o '") + imagemap + string("' '") + dotfile + string("'");
    retVal = system(cmd.c_str());
    if (retVal != 0) 
	goto error;
    retVal = system((string("dot -Tpng -o '") + callgraph + string("' '") + dotfile + string("'")).c_str());
    if (retVal != 0)
	goto error;

    out << "Callers for " << f->owningClass()->shortName() << "." << f->name()
	<< "<br><br>" << endl;
    out << "<img src=\"" << f->callgraphImageFilename()
	<< "\" usemap=\"#callgraph\" border=\"0\">" << endl;
    includeFile(out, imagemap);

    out.close();
    return true;

 error:
    out << "Error generating callgraph" << "</html>";
    out.close();
    return false;
}

bool JavaBrowser::writeFileIndex()
{
    ofstream out((_outputDir + kFileIndexFile).c_str());

    if (!out.is_open())
	return false;

    out.close();
    return true;
}

bool JavaBrowser::includeFile(ofstream& out, const string& file)
{
    string line;
    ifstream imap(file.c_str());
    if (imap.is_open()) {
	while (imap.good()) {
	    getline(imap, line);
	    out << line;
	}
	imap.close();
    }
    else
	return false;
    return true;
}

string JavaBrowser::getLink(const Name& type)
{
    Class& c = _ns.get(type);
    if (c.seenDefinition())
	return ::getClassLink(c, c.shortName());
    return html(type);
}

void JavaBrowser::writeHeader(ofstream& out, const std::string& title)
{
    out << "<html>" << endl;
    out << "<head>" << endl;
    out << "<title>" << title << "</title>" << endl;
    out << "<link href=\"" << kStyleSheetFile
	<< "\" rel=\"stylesheet\" type=\"text/css\">" << endl;
    out << "<script language=\"javascript\" src=\"" << kJavascriptFile
	<< "\"></script>" << endl;
    out << "</head>" << endl;
}

void JavaBrowser::writeStyleSheet()
{
    ofstream out((_outputDir + kStyleSheetFile).c_str());

    // Using nowrap prevents the tooltips from causing line wraps if the
    // line being hovered over happens to be longer than the viewport width.
    out << "body { font-family: sans-serif; white-space: nowrap; }" << endl;

    // Make an unbold.  Ordinarily you'd use a span, but since I'm planning
    // on putting these in the floating tooltip spans, this is easier
    // than figuring out how to undo absolute in relative positioning...
    out << "B.unbold { color: #444444;  font-weight: normal; }" << endl;

    out << "SPAN.keyword       { color: #333333; }" << endl;
    out << "SPAN.keywordtype   { color: #000077 }" << endl;
    out << "SPAN.keywordflow   { color: #000000; font-weight: bold; }"
         << endl;
    out << "SPAN.comment       { color: #005500 }" << endl;
    out << "SPAN.preprocessor  { color: #cc33cc }" << endl;
    out << "SPAN.string        { color: #440044 }" << endl;
    out << "A:link { text-decoration: none; }" << endl;
    out << "A:visited { text-decoration: none; }" << endl;
    out << "A:hover { background-color: #ffffee }" << endl;
    // Specifying position: relative means that position: absolute in the
    // the subtags (i.e. the tooltip) will be specified relative to the link
    // (This way it acts as a proper tooltip)
    out << "A.var:link { position: relative; text-decoration: none; font-weight: normal; color: #000000; }" << endl;
    out << "A.var:visited { position: relative; text-decoration: none; font-weight: normal; color: #000000; }" << endl;
    out << "A.func:link { position: relative; text-decoration: none; font-weight: normal; color: #0000ff; }" << endl;
    out << "A.func:visited { position: relative; text-decoration: none; font-weight: normal; color: #0000ff; }" << endl;
    out << "A.type:link { text-decoration: none; font-weight: normal; color: #0000ff; }" << endl;
    out << "A.type:visited { text-decoration: none; font-weight: normal; color: #0000ff; }" << endl;
    out << "A.define:link { text-decoration: none; font-weight: normal; color: #cc33cc; }" << endl;
    out << "A.define:visited { text-decoration: none; font-weight: normal; color: #cc33cc; }" << endl;
    out << "TD.memberlist { text-align: right }" << endl;
    out << "TR.inherited { background-color: #eeeeee }" << endl;

    // Equivalent to border="0" cellpadding="0" cellspacing="0", but
    // allows class="func-tip", which is shorter.
    out << "TABLE.func-tip { border-collapse : collapse; }" << endl;
    out << "TABLE.func-tip td, TABLE.func-tip th { padding: 0; vertical-align: top; }" << endl;

    out << "TABLE.nav { background-color: #eeeeee; border-bottom: 1px solid; position: fixed; top: 0pt; bottom: auto; right: 0pt; left: 0pt; width: 100% }" << endl;
    out << "TABLE.rightnav { background-color: #eeeeee; border-bottom: 1px solid; border-left: 1px solid; position: fixed; top: 0pt; bottom: auto; right: 0pt; left: auto; }" << endl;
    out << "span { position: relative; }" << endl;
    out << "a span { display: none; }" << endl;
    out << "a:hover span { display: block; font-family: sans-serif; color: black; background-color: #ffffee; border: 1px solid; position: absolute; top: -30px; left: 75px; z-index: 1; }" << endl;
    out << "span span { display: none; }" << endl;
    out << "span:hover span { display: block; font-family: sans-serif; color: black; background-color: #ffffee; border: 1px solid; position: absolute; top: -30px; left: 75px; z-index: 1; }" << endl;

    out.close();
}

void JavaBrowser::writeJavascript()
{
    ofstream out((_outputDir + kJavascriptFile).c_str());

    out << "function load(classHref, srcHref, callerHref) {" << endl;
    out << "    if (classHref != \"\" && top.class.location != classHref)"
        << endl;
    out << "        top.class.location=classHref;" << endl;
    out << "    if (srcHref != \"\" && top.src.location != srcHref)" << endl;
    out << "        top.src.location=srcHref;" << endl;
    out << "    if (callerHref != \"\" && top.caller.location != callerHref)"
	<< endl;
    out << "        top.caller.location=callerHref;" << endl;
    out << "}" << endl;
    out << endl;
    out << "function toggleVisibilityLink(className, id){" << endl;
    out << "  toggleVisibility(document.body, className);" << endl;
    out << "  var node = document.getElementById(id);" << endl;
    out << "  if (node.innerHTML.search('Show') >= 0)" << endl;
    out << "    node.innerHTML = node.innerHTML.replace(/Show/, 'Hide');" << endl;
    out << "  else" << endl;
    out << "    node.innerHTML = node.innerHTML.replace(/Hide/, 'Show');" << endl;
    out << "}" << endl;
    out << endl;
    out << "function toggleVisibility(node, className){" << endl;
    out << "  if (node.className == className) {" << endl;
    out << "    if (node.style.display == 'none')" << endl;
    out << "      node.style.display = '';  // 'block' messes up tables in firefox" << endl;
    out << "    else" << endl;
    out << "      node.style.display = 'none';" << endl;
    out << "  } else if (node.firstChild)" << endl;
    out << "    toggleVisibility(node.firstChild, className);" << endl;
    out << "  if (node.nextSibling)" << endl;
    out << "    toggleVisibility(node.nextSibling, className);" << endl;
    out << "}" << endl;

    out.close();
}
