//-------------------- Copyright (C) 2009 by Geoffrey Prewett -----------------
#include <stdio.h>	// For file operations;  printf, etc. not used
#include <stdlib.h>	// For atoi()
#include <assert.h>

#include <iostream>
#include <string>
#include <map>
#include <set>
#include <vector>
#include <sstream>

#include "Global.h"
#include "Exception.h"
#include "Counter.h"
#include "Namespace.h"
#include "ScopedNames.h"
#include "Graph.h"
#include "GraphParseOutput.h"
#include "JavaParser.h"

using namespace std;

void printUsage(char *progName)
{
    cout << "Usage: " << progName << " [options] file1.java [file2.java ... fileN.java]" << endl;
    cout << "    --debug\t\t\t" << "Outputs copious parsing information" << endl;
    cout << "    --min-edges n\t\t" << "Outputs nodes that have >= n inputs" << endl;
    cout << "    --only-show-bidirectional-nodes\t" << "Outputs only nodes that have" << endl << "\t\t\t\tboth inputs and outputs" << endl;
    cout << "    --only-show-cycles\t\t" << "Outputs nodes in cycles" << endl;
    cout << "    --show-hierarchy\t\t"<< "Draws bold edges indicating class hierarchy" << endl;
    cout << "    --show-undefined\t\t" << "Outputs classes that are called, but not defined" << endl;
    cout << "    --title str\t\t" << "Draws title text.  Use '\n' to separate multiple lines" << endl;
}

Parser* parserFactory(const char *filename, Namespace &ns, GraphParseOutput &o)
{
    string file(filename);
    string type = file.substr(file.rfind(".") + 1);

    if (type == "java")
	return new JavaParser(ns, o, filename);

    return NULL;
}

int main(int argc, char *argv[])
{
    Namespace nameSpace;
    Graph graph(nameSpace);
    GraphParseOutput output(graph);

    if (argc < 2) {
	printUsage(argv[0]);
	return 0;
    }

    gSettings.debug = false;
    gSettings.minEdges = 1;
    gSettings.noEndOrStartNodes = false;
    gSettings.onlyShowCycles = false;
    gSettings.showHierarchy = false;
    gSettings.showUndefinedClasses = false;
    gSettings.title += argv[0];
    for (int i = 0;  i < argc;  ++i) {
	if (argv[i][0] == '-' && argv[i][1] == '-')
	    gSettings.title += string(" ") + string(argv[i]);
    }
    gSettings.tabSize = 8;

    int start = 1;
    while (start < argc && argv[start][0] == '-' && argv[start][1] == '-') {
	if (string(argv[start]) == "--debug")
	    gSettings.debug = true;
	else if (string(argv[start]) == "--min-edges") {
	    gSettings.minEdges = atoi(argv[start + 1]);
	    start++;
	}
	else if (string(argv[start]) == "--only-show-bidirectional-nodes")
	    gSettings.noEndOrStartNodes = true;
	else if (string(argv[start]) == "--only-show-cycles")
	    gSettings.onlyShowCycles = true;
	else if (string(argv[start]) == "--show-hierarchy")
	    gSettings.showHierarchy = true;
	else if (string(argv[start]) == "--show-undefined")
	    gSettings.showUndefinedClasses = true;
	else if (string(argv[start]) == "--title") {
	    gSettings.title = argv[start + 1];
	    start++;
	}
	else {
	    printUsage(argv[0]);
	    return 0;
	}
	start++;
    }
    gSettings.title += "\\n";  // Reserve space below title

    for (int i = start;  i < argc;  ++i) {
	try {
	    Parser *p = parserFactory(argv[i], nameSpace, output);
	    if (!p) {
		cerr << "Could not find a parser for " << argv[i] << endl;
		continue;
	    }
	    if (gSettings.debug)
	      cerr << "--------------- " << argv[i] << " (header) -------------"
		   << endl;
	    p->parse(Parser::kBuildHierarchy);
	} catch(Exception& e) {
	    cerr << e.msg() << endl;
//	} catch(...) {
//	    cerr << gParseState.file << ":" << gParseState.lineNo
//		 << ":  Internal error:  unhandled exception" << endl;
	}
    }

    if (gSettings.debug)
	nameSpace.print();

    for (int i = start;  i < argc;  ++i) {
	try {
	    Parser *p = parserFactory(argv[i], nameSpace, output);
	    if (!p)  // We already spit out a warning, no need to do it again
		continue;
	    if (gSettings.debug)
	      cerr << "--------------- " << argv[i] << " (parse) -------------"
		   << endl;
	    p->parse(Parser::kOutputCallGraph);
	} catch(Exception& e) {
	    cerr << e.msg() << endl;
	} catch(...) {
	    cerr << gParseState.file << ":" << gParseState.lineNo
		 << ":  Internal error:  unhandled exception" << endl;
	}
    }
//    gNamespace.printSubGraphs();
//    gGraph.print();
    graph.printClasses();

    return 1;
}
