//-------------------- Copyright (C) 2009 by Geoffrey Prewett -----------------
#include "JavaParser.h"

bool JavaParser::parseLine(ScopedNames& scope, TokenArray& tokens, Mode mode)
{
    unsigned int idx = nextToken(tokens, -1); // Get first non-whitespace token
    if (idx >= tokens.size())  // Line with only whitespace and/or comments
	return true;

    if (tokens[idx] == "package") {
	tokens[idx].setType(Token::kKeyword);
	unsigned int i = nextToken(tokens, idx);
	std::string pkg;
	while (i < tokens.size() && tokens[i] != ";") {
	    if (tokens[i].type() != Token::kWhitespace)
		pkg += tokens[i].str();
	    i = nextToken(tokens, i);
	}
	_nameResolutionList.push_back(NameLookupInfo(pkg, false));
	scope.setPackage(pkg);
	return true;
    }
    if (tokens[idx] == "import") {
	tokens[idx].setType(Token::kKeyword);

	// Java 1.5 static import (imports static functions into the namespace)
	unsigned int next = nextToken(tokens, idx);
	if (next < tokens.size() && tokens[next] == "static") {
	    tokens[next].setType(Token::kKeyword);
	    idx = next;
	}

	std::string import;
	idx = joinTokens(tokens, nextToken(tokens, idx), ".", &import);
	assert(idx < tokens.size());
	if (import[import.size() - 1] == '*') {
	    import.erase(import.size() - 2, 2);
	    _nameResolutionList.push_back(NameLookupInfo(import, false));
	} else
	    _nameResolutionList.push_back(NameLookupInfo(import, true));
	return true;
    }

    // Process annotations
    while (tokens[idx] == "@") {
	Name annotation;
	unsigned int next = nextToken(tokens, idx);
	unsigned int end = joinTokens(tokens, next, ".", &annotation);
	tokens[idx].setType(Token::kKeyword);
	tokens[next].setStr(annotation);
	if (annotation == "Deprecated" || annotation == "Override") {
	    tokens[next].setType(Token::kKeyword);
	    idx = nextToken(tokens, next);
	} else {
	    Name annotationClass;
	    if (!resolveClass(scope, annotation, &annotationClass))
		annotationClass = annotation;
	    tokens[next].setType(Token::kClass);
	    tokens[next].setClassRef(annotationClass);

	    if (end < tokens.size() && tokens[end] == "(") {
		next = end;  // This should be the '('
		idx = nextToken(tokens, findMatchingParen(tokens, next));
		tokens[idx].setType(Token::kKeyword);
		while (idx < next) { // This annotation stretches across multiple lines
		    appendNextLine(&tokens);
		    idx = nextToken(tokens, findMatchingParen(tokens, next));
		}
	    } else {
		idx = nextToken(tokens, next);
	    }
	}
    }

    // Process scoping info (public, private, static, etc.)
    int modifiers = 0;
    int tokenModifier = 0;
    while (idx < tokens.size() &&
	   (tokenModifier = Member::parseModifier(tokens[idx].str()))) {
	modifiers |= tokenModifier;
	tokens[idx].setType(Token::kKeyword);
	idx = nextToken(tokens, idx);
    }

    // At this point we either have a class definition, a method or
    // variable definition, a block open or close, or a normal expression.

    if (tokens[idx] == "class" || tokens[idx] == "interface" ||
	(idx + 1 < tokens.size() && tokens[idx] == "@" && tokens[idx+1] == "interface")) {
	assert(idx + 1 < tokens.size());
	if (tokens[idx] == "@")
	    tokens[idx++].setType(Token::kKeyword);
	tokens[idx].setType(Token::kKeyword);
	Name owningClass = scope.getClass();

	Name templatedName = tokens[nextToken(tokens, idx)].str();
	std::string::size_type pos = templatedName.find('<');
	Name className = templatedName.substr(0, pos);
	Name templateType;
	if (pos != std::string::npos) {
	    templateType = templatedName.substr(pos + 1);
	    templateType = templateType.substr(0, templateType.find(' '));
	}

	scope.pushClass(className);
	// Add this class into the global namespace
	if (mode == kBuildHierarchy) {
	    unsigned int idx2 = nextToken(tokens, idx);
	    Name className = scope.getClass();
	    Class &info = _namespace.get(className);
//	    info.setName(className);
	    if (!templateType.empty())
		info.setTemplateParameter(templateType);
	    info.setDefinition(gParseState.file, gParseState.lineNo);
	    if (tokens[idx] == "interface")
		info.setInterface();
	    info.setShortName(className.substr(scope.package().size() + 1));
	    if (scope.depth() >= 2) {
		info.setOuterClass(owningClass);
		_namespace.get(owningClass).addInnerClass(className);
	    }

	    while (idx2 < tokens.size() && tokens[idx2] != "{")
	    {
		if (tokens[idx2] == "extends" || tokens[idx2] == "implements" ||
		    tokens[idx2] == ",")
		{
		    Name parent;
		    bool concrete = (tokens[idx2] == "extends");
		    idx2 = joinTokens(tokens, nextToken(tokens, idx2), ".",
				      &parent);
		    tokens[idx2].setClassRef(parent);
		    Name fullParent;
		    if (!resolveClass(scope, parent, &fullParent))
			fullParent = parent;
		    info.addParent(fullParent);
		    _namespace.get(fullParent).addChild(className);
		    // Since we know a bit of information about the parent
		    // class, record it in case we never encounter it
		    if (!concrete &&
			!_namespace.get(fullParent).seenDefinition())
			_namespace.get(fullParent).setInterface();
		}
		idx2 = nextToken(tokens, idx2);
	    }
	}
	return true;
    }
    if (tokens[idx] == "{") {
	scope.pushBlock();
	return true;
    }
    if (tokens[idx] == "}") {
	scope.popBlock();
	return true;
    }

    // Now we either have a function declaration or a normal expression

    // Process constructor (special case of function definition)
    Name currentClass = scope.getClass();
    if (idx + 1 < tokens.size() &&
	// The shortName() of the class is the name without the package,
	// but if this is an inner class, we just want the last part
	tokens[idx] == currentClass.substr(currentClass.rfind('.') + 1) &&
	tokens[nextToken(tokens, idx)] == "(" &&
	tokens.back() != ";" /* if ;, then is a member var ref to self */) {
	Name func = tokens[idx].str();
	idx = nextToken(tokens, idx, 2);
	Function::Args args;
	idx = processFunctionArgs(tokens, idx, scope, &args);
	if (mode == kBuildHierarchy)
	    this->addName(scope, kMethod, func, "", modifiers, args);
	scope.pushMethod(func);
	for (unsigned int argIdx = 0;  argIdx < args.size();  ++argIdx)
	    scope.add(args[argIdx].name, args[argIdx].type);
	return true;
    }
    // Prevent these from looking like a function declaration
    if (tokens[idx] == "return" || tokens[idx] == "for" ||
	tokens[idx] == "if" || tokens[idx] == "throw") {
	tokens[idx].setType(Token::kKeyword);
	idx = nextToken(tokens, idx);
    }

    if (tokens[idx] == "new" || tokens[idx] == "else")
	tokens[idx].setType(Token::kKeyword);

    if (idx <= tokens.size() - 3 && tokens[idx] != "new" &&
	tokens[idx] != "else" && // "else if (...)" is not a function...
	isName(tokens[idx].str()) &&
	isName(tokens[nextToken(tokens, idx)].str())) {
	if (tokens[nextToken(tokens, idx, 2)] == "(") { // Function declaration
	    Name retType(tokens[idx].str());
	    Name fullRetType;
	    if (isBuiltinType(retType))
		tokens[idx].setType(Token::kBuiltinType);
	    else if (resolveClass(scope, retType, &fullRetType))
		tokens[idx].setClassRef(fullRetType);
	    if (fullRetType.empty()) // i.e. resolveClass() failed
		fullRetType = retType;
	    assert(!fullRetType.empty());
	    idx = nextToken(tokens, idx);
	    const Name& func = tokens[idx].str();
	    tokens[idx].setDefinition(Token::kFunctionDefinition,
				      scope.getMethod());

	    // Process the parameters
	    Function::Args args;
	    idx = nextToken(tokens, idx, 2);
	    idx = processFunctionArgs(tokens, idx, scope, &args);

	    if (mode == kBuildHierarchy) {
		modifiers |= Member::kVirtual;  // everything in Java is virtual
		this->addName(scope, kMethod, func, fullRetType, modifiers,
			      args);
	    }

	    // If this is not an abstract method it will have a body
	    // so push the scope
	    if (tokens.back() == "{") {  // if abstract, will be ";"
		scope.pushMethod(func);
		for (unsigned int argIdx = 0;  argIdx < args.size();  ++argIdx)
		    scope.add(args[argIdx].name, args[argIdx].type);
	    } else {  assert(tokens.back() == ";"); }

	    return true;
	} else if (tokens[nextToken(tokens, idx, 2)] == ";" ||
	       tokens[nextToken(tokens, idx, 2)] == "=") {//Variable declaration
	    Name retType(tokens[idx].str());
	    Name fullName;
	    if (isBuiltinType(tokens[idx].str()))
		tokens[idx].setType(Token::kBuiltinType);
	    else if (resolveClass(scope, tokens[idx].str(), &fullName))
		tokens[idx].setClassRef(fullName);
	    idx = nextToken(tokens, idx);
	    const Name& var = tokens[idx].str();
	    tokens[idx].setDefinition(Token::kObjectDefinition,
				      scope.getMethod());
	    if (!scope.inMethod() && mode == kBuildHierarchy)
		this->addName(scope, kVariable, var, retType, modifiers);
	    scope.add(var, retType);
	    idx = nextToken(tokens, idx, 2);// Process any assignment expression
	}
    }

    if (mode == kOutputCallGraph) {
	this->parseExpression(tokens, idx, scope);
    }
    else {
	while (idx < tokens.size()) {
	    if (tokens[idx] == "{")  // generally if/for/while
		scope.pushBlock();
	    else if (tokens[idx] == "}")
		scope.popBlock();
	    idx = nextToken(tokens, idx);
	}
    }

    return true;
}

unsigned int JavaParser::processFunctionArgs(TokenArray& tokens,
					     unsigned int idx,
					     ScopedNames& scope,
					     Function::Args *args)
{
    while (idx < tokens.size() && tokens[idx] != ")") {
	int flags = (Function::Arg::Flags)0;
/*	if (tokens[idx] == "const") {
	    flags |= Function::Arg::kConst;
	    tokens[idx].setType(Token::kKeyword);
	    idx = nextToken(tokens, idx);
	} */
	if (tokens[idx] == "final") {
	    flags |= Function::Arg::kFinal;
	    tokens[idx].setType(Token::kKeyword);
	    idx = nextToken(tokens, idx);
	}
	std::string argType, argName, defaultVal;
	int typeIdx = idx;
	idx = joinTokens(tokens, idx, ".", &argType);
	idx = nextToken(tokens, idx);
	this->setClassToken(tokens[typeIdx], tokens[typeIdx].str(), scope);
/*	if (tokens[idx] == "const") {
	    flags |= Function::Arg::kConst;
	    tokens[idx].setType(Token::kKeyword);
	    idx = nextToken(tokens, idx);
	} */
	argName = tokens[idx].str();
	if (argName == "...") {
	    argType += "[]";
	    idx = nextToken(tokens, idx);
	    argName = tokens[idx].str();
	}
	idx = nextToken(tokens, idx);
	if (idx < tokens.size() && tokens[idx] == "=") {
	    while (tokens[idx] != "," && tokens[idx] != ")") {
		if (!defaultVal.empty())
		    defaultVal += " ";
		defaultVal += tokens[idx].str();
		idx = nextToken(tokens, idx);
	    }

	}
	args->push_back(Function::Arg(argName, argType, flags,
				      defaultVal));
	assert(tokens[idx] == "," || tokens[idx] == ")");
	if (tokens[idx] == ",")
	    idx = nextToken(tokens, idx);
    }

    return idx;
}

bool JavaParser::isObjDereference(const std::string& token)
{
    return (token == ".");
}

bool JavaParser::isBuiltinType(const std::string& token)
{
    if (token == "void" || token == "boolean" || token == "byte" ||
	token == "char" || token == "short" || token == "int" ||
	token == "long" || token == "float" || token == "double")
	return true;
    return false;
}

/*
State JavaParser::nextState(TokenArray& tokens, int idx, State state)
{
    if (tokens[idx] == ".")
	return kObjDeref;
    else if (tokens[idx] == "{") {
	return kOpenBlock;
    }
    else if (tokens[idx] == "}")
	return kCloseBlock;
    else if (tokens[idx] == "@")
	return kAnnotation;
    else if (tokens[idx] == "static" || tokens[idx] == "public" ||
	     tokens[idx] == "private" || tokens[idx] == "protected" ||
	     tokens[idx] == "final" || tokens[idx] == "abstract")
	return kScopeModifier;
    else if (tokens[idx] == "import")
	return kImport;
    else if (tokens[idx] == "package")
	return kPackage;
    else
	return kUnimportant;
}
*/
