//-------------------- Copyright (C) 2009 by Geoffrey Prewett -----------------
#include "Parser.h"
#include "Exception.h"
#include "Function.h"

#include <iostream>

using namespace std;

int gDummyVar = 0;
bool Parser::parse(Mode mode)
{
    ScopedNames scope(_namespace);

    if (gSettings.debug)
	cerr << "---- Beginning parsing of " << _filename << " -----" << endl;

    _in = fopen(_filename.c_str(), "r");
    if (!_in)
	throw Exception((string("Could not read ") + _filename));
    gParseState.file = _filename;
    gParseState.lineNo = 1;
    gParseState.actualLineNo = 1;

    _output.start();

    string line;
    while (nextLine(&line)) {
	if (gParseState.lineNo == 54)
	    line = line;
	TokenArray tokens = getTokens(line);

	if (gSettings.debug) {
	    cerr << gParseState.lineNo << ":  ";
	    printTokenArray(tokens);
	}

	if (tokens.empty())
	    continue;

	this->parseLine(scope, tokens, mode);
 
	TokenArray::iterator it;
//    cerr << "Tokens:" << endl;
	for (it = tokens.begin();  it != tokens.end();  ++it) {
//	cerr << "   " << it->type() << "  '" << it->str() << "'" << endl;
	    switch (it->type()) {
	 	case Token::kWhitespace:
		    _output.whitespaceToken(it->str());
		    break;
		case Token::kComment:
		    _output.commentToken(it->str());
		    break;
		case Token::kUnknown:
		case Token::kNothing:
		    _output.otherToken(it->str());
		    break;
		case Token::kKeyword:
		    _output.keywordToken(it->str());
		    break;
	        case Token::kConstant:
		case Token::kString:
		    _output.stringToken(it->str());
		    break;
		case Token::kObject:
		{
		    Symbol *sym = NULL;
//		    cout << "Outputting kObject token: '" << it->str() << "'" << endl;
		    if (it->str() == "super") {
			Name parent = _namespace.get(scope.getClass()).concreteParent(_namespace);
			if (!parent.empty())
			    sym = &_namespace.get(parent);
		    }
		    else {
			Name name;
			if (resolveClass(scope, it->str(), &name)) {
			    // See if this is really a class
			    if (_namespace.has(name)) {
//				cout << "   found class: " << _namespace.get(name).name() << endl;
				_output.classToken(&_namespace.get(name),
						   it->str());
				break;
			    }
			}

			// See if this is a member variable
			name = scope.getScopedName(it->str());
			std::string::size_type pos = name.rfind(".");
			Name className = name.substr(0, pos);
			Name memberName = name.substr(pos + 1);
//			cout << "   class: " << className << ", member: " << memberName << endl;
			Name actualClass;
			Member *member;
			if (_namespace.lookup(className, memberName,
					      Symbol::kMember,
					      &actualClass, &member)) {
//			    cout << "   found member!" << endl;
			    sym = member;
			} else {
			    // Maybe it's in the scope somewhere
			    // (Note:  this won't work if the scope shadows
			    //	       a member variable!)
			    Name type = scope.find(it->str());
			    if (type.empty()) {
				// If we are in a method, it might be in the
				// method's arguments
				if (scope.inMethod()) {
				    Name method = scope.getMethod();
				    pos = name.rfind(".");
				    className = method.substr(0, pos);
				    memberName = method.substr(pos + 1);
//				    cout << "   not a member, but looking through args of: " << className << " . " << memberName << endl;
				    if (_namespace.lookup(className, memberName,
							  Symbol::kMember,
						       &actualClass, &member)) {
					if (member->symbolType() == Symbol::kFunction)
					{
					    Function *f = (Function *)member;
					    Function::Args::const_iterator arg;
					    for (arg = f->args().begin();
						 arg != f->args().end(); ++arg){
//						cout << "      Arg: " << arg->name << " (" << arg->type << ")" << endl;
						if (arg->name == it->str()) {
						    sym = &_namespace.get(arg->type);
						    break;
						}
					    }
					}
				    }
				}
			    }
			    else {
//				cout << "   not a member, but found type: "
//				     << type << endl;
				sym = &_namespace.get(type);
			    }
			}
		    }
		    assert(!sym || !sym->name().empty());
		    _output.objectToken(it->str(), sym);
//		    cout << "-----" << endl;
		    break;
		}
		case Token::kBuiltinType:
		    _output.builtinTypeToken(it->str());
		    break;
		case Token::kClass:
		    _output.classToken(&_namespace.get(it->fullClassName()),
				       it->str());
		    break;
		case Token::kOperator:
		    _output.operatorToken(it->str());
		    break;
		case Token::kFunctionCall:
		    _output.functionCall(it->caller(), it->callee(),
					 it->callThru(), it->str());
		    break;
		case Token::kFunctionDefinition:
		    _output.functionDefinition(it->defScope(), it->str());
		    break;
		case Token::kObjectDefinition:
		    _output.objectDefinition(it->defScope(), it->str());
		    break;
	    }
	}
    }

    if (scope.depth() != 0)
	cerr << gParseState.file << ":" << gParseState.actualLineNo
	     << ":  Internal error:  scope.depth() != 0 at "
	     << __FILE__ << ":" << __LINE__ << endl;

    fclose(_in);
    return true;
}

void Parser::addName(ScopedNames& scope, NameType type,
		     const Name& templatedName, const Name& retType,
		     int modifiers,
		     const Function::Args& args /*= Function::Args()*/
    		     )
{
    Name name = untemplatedName(templatedName);
    Name thisClass = scope.getClass();

    if (type == kMethod) {
	Name actualClass;
//	Member *existingMember;
//	bool exists = _namespace.lookup(thisClass, name,
//					&actualClass, &existingMember);
//
//	if (exists && /*existingType.isValid() &&*/
//	    existingMember->type() != retType)
//	    // (if type is equal, is an overloaded function, so is ok)
//	    cerr << gParseState.file << ":" << gParseState.lineNo 
//		 << ":  redefining method name " << thisClass << "."
//		 << name <<" with type " << retType << endl;
	if (gSettings.debug)
	    cerr << "Method:  " << thisClass << "." << name << " -> "
		 << retType << endl;
	Class& c = _namespace.get(thisClass);
	Function *f = new Function(&c, name, retType, args, modifiers);
	f->setDefinition(gParseState.file, gParseState.lineNo);
	c.members().insert(name, f);
    } else if (type == kVariable) {
	if (!scope.inMethod()) {  // Member variable
	    Name actualClass;
//	    Member *existingMember;
//	    bool exists = _namespace.lookup(thisClass, name,
//					    &actualClass, &existingMember);
//	    if (exists)
//		cerr << gParseState.file << ":" << gParseState.lineNo
//		     << ":  redefining variable name " << thisClass
//		     << "." << name << " with type "
//		     << retType << endl;
	    if (gSettings.debug)
		cerr << "Variable:  " << thisClass << "." << name
		     << " -> " << retType << endl;
	    Class& c = _namespace.get(thisClass);
	    Member *m = new Member(&c, name, retType, modifiers);
	    m->setDefinition(gParseState.file, gParseState.lineNo);
	    c.members().insert(name, m);
	}
    }
}

void Parser::parseExpression(TokenArray& tokens, unsigned int idx,
			     ScopedNames& scope)
{
    Name retType, lastObj, funcCast;
    const Name invalidFunc = "void";
    Name lastFunc = invalidFunc;
    bool lastOpenParenFromNew = false;

    if (gSettings.debug) {
	cerr << "   (debug) parseExpression():  ";
	printTokenArray(tokens);
    }
    while (idx < tokens.size()) {
	if (tokens[idx] == "new") {
	    // Three types of new:
	    //    new Object[] { ... }
	    //    new Object( ... )
	    //    new Object( ... ) { ... }   (anonymous class)

	    tokens[idx].setType(Token::kKeyword);
	    idx = nextToken(tokens, idx);

	    if (idx < tokens.size() &&
		tokens[idx].str().rfind("[]") != string::npos) { // type 1 (array)
		setClassToken(tokens[idx], tokens[idx].str(), scope);
		idx = nextToken(tokens, idx);
		continue;
	    }
	    // At this point we know we are either type 2 (object) or 3.
	    // Parse through any full names (e.g. Object.InnerClass)
	    string typeName;
	    while (idx < tokens.size() && tokens[idx] != "(") {
//		if (tokens[idx].type() == Token::kUnknown)
//		    tokens[idx].setType(Token::kNothing);
		typeName += tokens[idx].str();
		if (tokens[idx] == ".")
		    tokens[idx].setType(Token::kNothing);
		else
		    setClassToken(tokens[idx], typeName, scope);
		idx = nextToken(tokens, idx);
	    }
	    lastOpenParenFromNew = true;
	    if (idx >= tokens.size())
		continue;
	}
	if (tokens[idx] == "return" || tokens[idx] == "for" ||
	    tokens[idx] == "while" || tokens[idx] == "if" ||
	    tokens[idx] == "catch") {
	    tokens[idx].setType(Token::kKeyword);
	    idx = nextToken(tokens, idx);
	}
	unsigned next = nextToken(tokens, idx);
	if (next < tokens.size() &&
	    isName(tokens[idx].str()) && tokens[next] == "(") {
	    Name func = tokens[idx].str(); // don't use ref here or we will
					   // crash if need to read more lines
					   // and realloc tokens
	    if (!retType.empty() && lastFunc != invalidFunc) {
		// Are parsing func().func2()
//		assert(lastObj.isValid());
		assert(!lastObj.empty());
		tokens[idx].setFunctionCall(lastObj + string(".") + lastFunc,
			   _namespace.getInheritedMethod(retType, func), true);
	    } else if (!retType.empty() && lastFunc == invalidFunc) {
		// Are parsing obj.func();  call is from this method.
		tokens[idx].setFunctionCall(scope.getMethod(),
			   _namespace.getInheritedMethod(retType, func), false);
	    } else { // Are parsing func()
		assert(tokens[idx].caller().empty());
		tokens[idx].setFunctionCall(scope.getMethod(),
					    scope.getScopedName(func), false);
	    }

	    if (gSettings.debug) {
		cerr << "   (debug) calling: " << func << "  (retType: "
		     << (retType.empty() ? "invalid" : retType )
		     << ", lastFunc: " << lastFunc << ")" << endl;
		cerr << "           recording call to " << tokens[idx].callee()
		     << endl;
	    }
	    idx = nextToken(tokens, idx);  // This is the "("
	    lastOpenParenFromNew = false;
	    unsigned int closeParen = findMatchingParen(tokens, idx);
	    while (closeParen <= idx) {  // must be "func( new ... { "
		appendNextLine(&tokens);
		closeParen = findMatchingParen(tokens, idx);
	    }
	    if (closeParen > nextToken(tokens, idx)){
		TokenArray subExpr;
		idx = nextToken(tokens, idx);
		unsigned int start = idx;
		while (idx < closeParen)
		    subExpr.push_back(tokens[idx++]);
		parseExpression(subExpr, 0, scope);
		// Copy any changes to the token type
		unsigned int ii = start;
		while (ii < closeParen) {
		    assert(subExpr[ii - start].str() == tokens[ii].str());
		    if (subExpr[ii - start].type() != tokens[ii].type())
			tokens[ii] = subExpr[ii - start];
		    ii++;
		}
	    }
	    if (closeParen > idx)
		idx = closeParen;
//	    lastObj = (retType.empty() ? scope.getClass() : retType);
	    if (!resolveClass(scope,
			      (retType.empty() ? scope.getClass() : retType),
			      &lastObj)) {
		// Might be a fully qualified name, or might not have been
		// defined
		lastObj = (retType.empty() ? scope.getClass() : retType);
	    }
//	    if (!funcCast.empty()) {
//		lastObj = funcCast;
//		retType = funcCast;
//		if (gSettings.debug)
//		    cerr << "   (debug) casting return type of function to '"
//			 << retType << "'" << endl;
//		funcCast = Name();
//	    }
	    lastFunc = func;
	} else if (this->isObjDereference(tokens[idx].str())) {
	    if (idx >= 1 && tokens[prevToken(tokens, idx)] == ")") {
						// have func().token[idx+1]
		Name obj;
		if (!retType.empty())  // had a type from a previous func/obj
		    obj = retType;
		else
		    obj = scope.getClass();
		// TODO:  what if you're in a inner class that references
		//        the outer class?
		if (gSettings.debug)
		    cerr << "   (debug)  func().  looking up '" << obj << "."
			 << lastFunc << "'";
		Member *member;
		if (!funcCast.empty()) {
		    lastObj = funcCast;
		    retType = funcCast;
		    if (gSettings.debug)
			cerr << "   (debug) casting return type of function to "
			     << "'" << retType << "'" << endl;
		}
		else if (_namespace.lookup(obj, lastFunc, Symbol::kFunction,
					   &obj, &member)) {
		    retType = member->type();
		    if (!resolveClass(scope, retType, &lastObj)) {
			// Might be a fully qualified name, or might not have
			// been defined
			lastObj = retType;
		    }
		}
		else if (_namespace.lookup(obj, lastFunc, Symbol::kMember,
					   &obj, &member)) {
		    retType = member->type();
		    if (!resolveClass(scope, retType, &lastObj)) {
			// Might be a fully qualified name, or might not have
			// been defined
			lastObj = retType;
		    }
		}
		else
		    retType = "";
		if (gSettings.debug)
		    cerr << " -> "
			 << ((retType.empty()) ? "invalid" : retType)
			 << endl;
	    } else {   // have   obj.token[idx+1]
		unsigned int prev = prevToken(tokens, idx);
		unsigned int next = nextToken(tokens, idx);
		if (next < tokens.size() && tokens[next] == "<") {
		    // have  obj.<type>token
		    next = nextToken(tokens, next);
		    funcCast = scope.find(tokens[next].str());
		    tokens[next].setType(Token::kObject);
		    if (gSettings.debug)
			cout << "   (debug)  found obj.<type> Java-function "
			     << "cast: type '" << tokens[next].str() << "' -> "
			     << funcCast << endl;
		    while (idx < tokens.size() && tokens[idx] != ">")
			idx = nextToken(tokens, idx);
		}
		if (gSettings.debug)
		    cerr << "   (debug)  obj.  lookup up '"
			 << tokens[prev].str() << "'";
		retType = scope.find(tokens[prev].str());
		if (retType.empty()) {
		    Name obj = retType;
		    Member *unused;
		    if (!_namespace.lookup(obj, tokens[prev].str(),
					   Symbol::kMember, &obj, &unused)){
			retType = tokens[prev].str();
			if (gSettings.debug)
			    cerr << " -> " << tokens[prev].str()
				 << " (assuming is a Java type) " << endl;
		    } else {
			if (gSettings.debug)
			    cerr << " -> "<< retType << " (found in "
				 << obj << ")" << endl;
		    }
		} else {
		  Name fullName;
		  if (resolveClass(scope, retType, &fullName))
		      retType = fullName;
		  if (gSettings.debug)
		      cerr << " -> " << retType << " (found in scope)"
			   << endl;
		}
	    }
	} else {
	    retType = Name();
	    lastObj = Name();
	    funcCast = Name();
	    lastFunc = invalidFunc;

	    if (tokens[idx] == "{") {  // generally if/for/while
		if (lastOpenParenFromNew) {
		    scope.pushClass("anonymous");
		}
		else
		    scope.pushBlock();
	    }
	    else if (tokens[idx] == "}")
		scope.popBlock();
	    else if (tokens[idx].type() == Token::kUnknown) {
		char c = tokens[idx].str()[0];
		if ((c >= 'a' && c <= 'z') ||
		    (c >= 'A' && c <= 'Z') || c == '_') {
		    if (tokens[idx] == "continue" || tokens[idx] == "break" ||
			tokens[idx] == "instanceof" || tokens[idx] == "try")
			tokens[idx].setType(Token::kKeyword);
		    else
			tokens[idx].setType(Token::kObject);
		}
	    }
	}

	idx = nextToken(tokens, idx);
    }
}

bool Parser::setClassToken(Token& token, const Name& className,
			   ScopedNames& scope)
{
    Name fullClassName;
    if (isBuiltinType(className)) {
	token.setType(Token::kBuiltinType);
	return true;
    }
    token.setType(Token::kClass);
    if (resolveClass(scope, className, &fullClassName)) {
	token.setClassRef(fullClassName);
	return true;
    }
    return false;
}

bool Parser::resolveClass(const ScopedNames& scope,
			  const Name& templatedClassName,
			  Name *resolvedName)
{
    Name className = untemplatedName(templatedClassName);
    Name nameIfInThisClass = scope.getClass() + "." + className;
    NameResolutionList::iterator it = _nameResolutionList.begin();
    for ( ;  it != _nameResolutionList.end();  ++it) {
	if (it->isClass && it->name == nameIfInThisClass) {
	    *resolvedName = nameIfInThisClass;
	    return true;
	} else {
	    Name tmp = it->name + className;
	    if (_namespace.has(tmp)) {
	        *resolvedName = tmp;
	        return true;
	    }
	}
    }

    it = _nameResolutionList.begin();
    for ( ;  it != _nameResolutionList.end();  ++it) {
	if (it->isClass) {
	    std::string::size_type pos = it->name.rfind('.');
	    if (pos != std::string::npos &&
		it->name.substr(pos + 1) == className) {
		*resolvedName = it->name;
		return true;
	    }
	} else {
	    string n = it->name + string(".") + className;
	    if (_namespace.has(n)) {
		*resolvedName = n;
		return true;
	    }
	}
    }

    *resolvedName = "";
    return false;
}

std::string Parser::untemplatedName(const Name& templatedName)
{
    return templatedName.substr(0, templatedName.find('<'));
}

int Parser::joinTokens(TokenArray& tokens, unsigned int startIdx,
		       const string& joinStr, string *joined)
{
    if (startIdx < tokens.size())
	*joined = tokens[startIdx++].str();
    while (startIdx < tokens.size() && tokens[startIdx] == joinStr)
    {
	*joined += tokens[startIdx++].str();
	while (startIdx < tokens.size() &&
	       tokens[startIdx].type() == Token::kComment)
	    startIdx++;
	if (startIdx < tokens.size())
	    *joined += tokens[startIdx++].str();
    }

    return startIdx--;
}

bool Parser::nextLine(string *line)
{
    bool inComment = false;
    bool inDblQuotes = false;
    bool inSingleQuotes = false;
    char c = ' ', b;
    line->clear();
    gParseState.lineNo = gParseState.actualLineNo;
    do {
	b = c;
	c = getc(_in);
	if (c == '\n') {
	    gParseState.actualLineNo++;
	    if (!inComment)
		gParseState.lineNo = gParseState.actualLineNo;
	}
	if (c == '\\') {
	    (*line) += c;
	    char cc = getc(_in);
	    if (cc == '"' || cc == '\'') {
		(*line) += cc;
		continue;
	    }
	    c = cc;
	}
	if (c == '\'' && !inDblQuotes && !inComment)
	    inSingleQuotes = !inSingleQuotes;
	if (c == '\"' && !inSingleQuotes && !inComment)
	    inDblQuotes = !inDblQuotes;
//	if (b == '/' && c == '/') {
//	    if (!line->empty())  // Might have '/* ... // ... */'
//		line->erase(line->size() - 1);
//	    while ((c = getc(_in)) != EOF && c != '\n')
//		;
//	    b = ' ';
//	    if (c == '\n') {
//		gParseState.actualLineNo++;
//		gParseState.lineNo = gParseState.actualLineNo;
//	    }
//	}
	if (!inDblQuotes && b == '/' && c == '*') {
	    inComment = true;
//	    line->erase(line->size() - 1);
	}
	if (!inDblQuotes && b == '/' && c == '/') {
	    while (c != EOF && c != '\n' && c != '\r') {
		(*line) += c;
		c = getc(_in);
	    }
	}

//	if (!inComment)
	    (*line) += c;

	if (!inDblQuotes && b == '*' && c == '/')
	    inComment = false;
	if (!inComment && !inDblQuotes && !inSingleQuotes &&
	    (c == ';' || c == '{' || c == '}'))
	    return true;
    } while(c != EOF);

    return false;
}

// Assumes that whitespace and comments are already marked.
unsigned int Parser::nextToken(TokenArray& tokens, unsigned int idx,
			       unsigned int n /*= 1*/)
{
    for (unsigned int i = 0;  i < n;  ++i) {
    	idx++;
    	while (idx < tokens.size() && 
	       (tokens[idx].type() == Token::kWhitespace ||
		tokens[idx].type() == Token::kComment))
	    idx++;
    }

    return idx;
}

// Assumes that whitespace and comments are already marked.
unsigned int Parser::prevToken(TokenArray& tokens, unsigned int idx,
			       unsigned int n /*= 1*/)
{
    for (unsigned int i = n;  i > 0;  --i) {
    	idx--;
    	while (idx < tokens.size() && 
	       (tokens[idx].type() == Token::kWhitespace ||
		tokens[idx].type() == Token::kComment))
	    idx--;
    }

    return idx;
}

Parser::TokenArray Parser::getTokens(const std::string& line)
{
    unsigned int pos = 0, size = line.size();
    TokenArray tokens;

    do {
	string tok;
	// Collect whitespace into a token
	while (pos < size && isWhitespace(line[pos])) {
	    tok.push_back(line[pos]);
	    pos++;
	}
	if (!tok.empty())
	    tokens.push_back(Token(tok, Token::kWhitespace));
	tok.clear();

	// Collect consequtive alphanumerics
	while (pos < size && !isWhitespace(line[pos]) &&
	       isAlphaNumeric(line[pos])) {
	    tok.push_back(line[pos]);
	    pos++;
	}
	if (!tok.empty())
	    tokens.push_back(Token(tok, Token::kUnknown));
	tok.clear();

	if (pos + 1 < size && line[pos] == '/' && line[pos + 1] == '/') {
	    while (pos < size && line[pos] != '\n')
		tok.push_back(line[pos++]);
	    tokens.push_back(Token(tok, Token::kComment));
	    if (line[pos] == '\n')
		continue;
	    break;
	}
	else if (pos + 1 < size && line[pos] == '/' && line[pos + 1] == '*') {
	    while (pos < size) {
		if (pos > 0 && line[pos] == '/' && line[pos - 1] == '*'){
		    tok.push_back(line[pos++]);
		    break;
		}
		tok.push_back(line[pos++]);
	    }
	    tokens.push_back(Token(tok, Token::kComment));
	}
	else if (line[pos] == '\'' && (pos == 0 || line[pos-1] != '\\')) {
	    do {
		tok.push_back(line[pos]);
		if (pos < size && line[pos] == '\\')
		    tok.push_back(line[++pos]);
		pos++;
	    } while (pos < size && line[pos] != '\'');
	    if (pos < size)  // Push back the trailing quote
		tok.push_back(line[pos++]);
	    tokens.push_back(Token(tok, Token::kConstant));
	}
	else if (line[pos] == '\"' && (pos == 0 || line[pos-1] != '\\')) {
	    do {
		tok.push_back(line[pos]);
		if (pos < size && line[pos] == '\\')
		    tok.push_back(line[++pos]);
		pos++;
	    } while (pos < size && line[pos] != '\"');
	    if (pos < size)  // Push back the trailing double quote
		tok.push_back(line[pos++]);
	    tokens.push_back(Token(tok, Token::kString));
	}
	else if (!isWhitespace(line[pos])) {
	    tok += line[pos];
	    tokens.push_back(Token(tok, Token::kUnknown));
	    pos++;
	}
    } while(pos < size);

//    cerr << "Raw tokens:  ";
//    printTokenArray(tokens);

//    bool inDblQuote = false;
    TokenArray deGenericedTokens;  // This isn't an efficient way to do this...
    for (unsigned int i = 0;  i < tokens.size();  ++i) {
//	if (inDblQuote) {
//	    deGenericedTokens.back().appendToken(tokens[i].str());
//	    if (tokens[i] == "\"" && (i == 0 || tokens[i - 1] != "\\")) {
//		inDblQuote = false;
//		deGenericedTokens.back().setType(Token::kString);
//	    }
//	    continue;
//	}
	if (tokens[i] != "<") {
	    // These are C++ tokens, but it is safe to always combine them,
	    // since they have no other meanings in other languages.
	    if (i > 0 && tokens[i] == ">" && tokens[i - 1] == "-") {
		deGenericedTokens.back().appendToken(">");
		continue;
	    }
	    else if (i > 0 && tokens[i] == ":" && tokens[i - 1] == ":") {
		deGenericedTokens.back().appendToken(":");
		continue;
	    }

	    if (i > 0 && ((tokens[i] == "|" && tokens[i - 1] == "|") ||
			  (tokens[i] == "&" && tokens[i - 1] == "&"))) {
		deGenericedTokens.back().appendToken(tokens[i].str());
		continue;
	    }
		
	    // There might be multiple dimensions in the array:  double[][][]
	    while (i + 1 < tokens.size() && tokens[i] == "[" &&
		tokens[nextToken(tokens, i)] == "]") {
		// The only time [] is legal is as an array type (e.g. double[])
		while (deGenericedTokens.back().type() == Token::kWhitespace)
		    deGenericedTokens.pop_back();
		deGenericedTokens.back().appendToken("[]");
		i = nextToken(tokens, i, 1) + 1;
	    }

	    if (tokens[i].type() == Token::kUnknown) {
		char c = tokens[i].str()[0];
		if (c >= '0' && c <= '9') {
		    tokens[i].setType(Token::kConstant);
		    // If we had "23." as the previous token and another
		    // numeric token for tokens[i], we can concatenate them
		    if (!deGenericedTokens.empty()) {
			Token& lastToken = deGenericedTokens.back();
			if (lastToken.type() == Token::kConstant &&
			    lastToken.str()[lastToken.str().size()-1] == '.') {
			    lastToken.appendToken(tokens[i].str());
			    continue;
			}
		    }
		    // We might get something like ".004" as well
		    if (!deGenericedTokens.empty() &&
			deGenericedTokens.back() == ".") {
			deGenericedTokens.back().appendToken(tokens[i].str());
			continue;
		    }
		}
		else if (c == '!' || c == '%' || c == '&' || c == '*' ||
			 c == '+' || c == '-' || c == '/' || c == '^' ||
			 c == '>' || c == '<' || c == '?' || c == ':' ||
			 c == '|' || c == ',')
		    tokens[i].setType(Token::kOperator);
		else if (tokens[i] == "true" || tokens[i] == "false" ||
			 tokens[i] == "null")
		    tokens[i].setType(Token::kConstant);
	    }

	    if (tokens[i] == "=") {
		if (i > 0 && tokens[i - 1].type() == Token::kOperator) {
		    deGenericedTokens.back().appendToken("=");
		    continue;
		}
		tokens[i].setType(Token::kOperator);
	    }
	    // If the last token is numeric and this token is a ".",
	    // then we can combine the two, since the "." is obvious part
	    // of the number.
	    if (i > 0 && tokens[i] == "." &&
		tokens[i - 1].type() == Token::kConstant) {
		deGenericedTokens.back().appendToken(tokens[i].str());
		continue;
	    }

	    // Handle functions with variable arguments ("...")
	    if (i > 0 && i + 2 <= tokens.size() &&
		tokens[i] == "." && tokens[i+1] == "." && tokens[i+2] == ".") {
		tokens[i].appendToken("..");
		deGenericedTokens.push_back(tokens[i]);
		i += 2;  // The loop increment will add one, too
		continue;
	    }

	    deGenericedTokens.push_back(tokens[i]);
//	    if (tokens[i] == "\"" && (i == 0 || tokens[i - 1] != "\\"))
//		inDblQuote = true;
	} /*else if (i > 0 && i + 2 < tokens.size() && tokens[i + 2] == ">") {
	    // If previous token was ".", then this must be a cast to an
	    // upcoming function call (e.g. this.<OtherClass>get(...) )
	    if (deGenericedTokens.back() != ".")
		deGenericedTokens.back().appendToken(tokens[i].str() + tokens[i+1].str() + tokens[i + 2].str());
	    i += 2;
	  */
	else {
	    unsigned int j = i + 1;
	    int nestedTemplates = 0;
	    bool hasBooleanOp = false;
	    while (j < tokens.size() && (tokens[j] != ">" || nestedTemplates)) {
		if (j > 0 && ((tokens[j] == "|" && tokens[j - 1] == "|") ||
			      (tokens[j] == "&" && tokens[j - 1] == "&")))
		    hasBooleanOp = true;
		if (tokens[j] == "<")
		    nestedTemplates++;
		if (tokens[j] == ">")
		    nestedTemplates--;
		j++;
	    }
	    if (hasBooleanOp || j >= tokens.size())
		deGenericedTokens.push_back(tokens[i]);
	    else if (!deGenericedTokens.empty() &&
		     deGenericedTokens.back() != ".") {
		deGenericedTokens.back().appendToken(tokens[i].str());
		for (unsigned int k = i + 1;  k <= j;  ++k)
		    deGenericedTokens.back().appendToken(tokens[k].str());
		i = j;
	    // If previous token was ".", then this must be a cast to an
	    // upcoming function call (e.g. this.<OtherClass>get(...) )
	    } else
		deGenericedTokens.push_back(tokens[i]);
	    
	}
    }
    return deGenericedTokens;
}

unsigned int Parser::findMatchingParen(TokenArray tokens, unsigned int startIdx)
{
    if (tokens[startIdx] != "(")
	cerr << gParseState.file << ":" << gParseState.lineNo
	     << ":  Internal error:  tokens[startIdx=" << startIdx
	     << "] != \"(\" at " << __FILE__ << ":" << __LINE__ << endl
	     << "  (token is '" << tokens[startIdx].str() << "')" << endl;

    int parenNum = 0;
    do {
	if (tokens[startIdx] == "(")  parenNum++;
	else if (tokens[startIdx] == ")")  parenNum--;
	startIdx++;
    } while(parenNum > 0 && startIdx < tokens.size());

    if (startIdx >= tokens.size() && tokens.back() != ")")
	return 0;
    return startIdx - 1;
}

void Parser::appendNextLine(TokenArray *tokens)
{
    string line;
    TokenArray nextTokens;

    nextLine(&line);
    nextTokens = getTokens(line);
    for (TokenArray::iterator it = nextTokens.begin();  it != nextTokens.end();
	 								   ++it)
	tokens->push_back(*it);
}

void Parser::printTokenArray(const TokenArray& tokens)
{
    string str;
    for (TokenArray::const_iterator it = tokens.begin();
	 it != tokens.end(); ++it) {
	str = it->str();
	if (it->type() == Token::kWhitespace) {
	    for (unsigned int i = 0;  i < str.size();  ++i) {
		if (str[i] == '\n')
		    std::cerr << "\\n";
		else if (str[i] == '\r')
		    std::cerr << "\\r";
		else if (str[i] == '\t')
		    std::cerr << "\\t";
		else if (str[i] == ' ')
		    std::cerr << "\\s";
		else
		    std::cerr << "\\d" << (int)str[i];
	    }
	}
	else {
	    string::size_type pos = 0;
	    while ((pos = str.find(' ')) != string::npos)
		str.replace(pos, 1, "\\s");
	    std::cerr << str;
	}

	std::cerr << " ";
    }
    std::cerr << std::endl;
}
