//-------------------- Copyright (C) 2009 by Geoffrey Prewett -----------------
#include "ScopedNames.h"
#include "Global.h"
#include <assert.h>

#include <iostream>

using namespace std;

void ScopedNames::pushBlock()
{
    _scopes.push_back(Name2Type());
    _names.push_back("");
    _types.push_back(kBlock);
    if (gSettings.debug) cerr << "   pushBlock():  depth: " << depth() << endl;
}

void ScopedNames::pushClass(const std::string& name)
{
    _scopes.push_back(Name2Type());
    _names.push_back(name);
    _types.push_back(kClass);
    if (gSettings.debug)
	cerr << "   pushClass(" << name << "):  depth: " << depth() << endl;
}

void ScopedNames::ScopedNames::pushMethod(const std::string& name)
{
    _scopes.push_back(Name2Type());
    _names.push_back(name);
    _types.push_back(kMethod);
    if (gSettings.debug)
	cerr << "   pushMethod(" << name << "):  depth: " << depth() <<endl;
}

void ScopedNames::popBlock()
{
    assert(!_scopes.empty());
    _scopes.pop_back();
    _names.pop_back();
    _types.pop_back();
    if (gSettings.debug) cerr << "   popBlock():  depth: " << depth() << endl;
}

bool ScopedNames::inMethod() const
{
    std::vector<ScopeType>::const_reverse_iterator rit = _types.rbegin();
    while (rit != _types.rend() && *rit == kBlock)
	rit++;
    if (rit != _types.rend() && *rit == kMethod)
	return true;
    return false;
}

Name ScopedNames::find(const Name& name) const
{
    if (name == "this")
	return getClass();
    if (name == "super") {
	Name super = _namespace.get(getClass()).concreteParent(_namespace);
	if (super == "" && _namespace.get(getClass()).parents().empty())
	    cerr << "Warning:  " << getClass()
		 << " has no parent but calls super()." << endl
		 << "          This is probably an internal error." << endl;
	return super;
    }

    std::vector<Name2Type>::const_reverse_iterator it = _scopes.rbegin();
    for (;  it != _scopes.rend();  ++it) {
	const Name *type = it->find(name);
	if (type)
	    return *type;
    }

    Name packageScope = _package + "." + name;
    if (_namespace.has(packageScope))
	return packageScope;

    return "";
}

Name ScopedNames::getClass() const
{
    Name n;
    for (unsigned int i = 0;  i < _names.size();  ++i) {
	if (_types[i] == kClass) {
	    if (!n.empty())
		n += ".";
	    n += _names[i];
	}
    }
    if (!n.empty() && !_package.empty())
	n = _package + std::string(".") + n;
    return n;
}

Name ScopedNames::getMethod() const
{
    Name method;
    for (unsigned int i = 0;  i < _names.size();  ++i) {
	if (_types[i] == kClass || _types[i] == kMethod) {
	    if (!method.empty())
		method += ".";
	    method += _names[i];
	}
    }

    if (!method.empty() && !_package.empty())
	method = _package + std::string(".") + method;
    return method;
}

Name ScopedNames::getScopedName(const Name& name) const
{
    if (name == "super") {  // This is super(), the constructor
	Name super = _namespace.get(getClass()).concreteParent(_namespace);
	if (super != "")
	    return super + string(".") + super.substr(super.rfind('.') + 1);
	return "<parent>.constructor";
    }

    // Figure out how many classes we're nested in
    int nClasses = 0;
    for (unsigned int i = 0;  i < _names.size();  ++i)
	if (_types[i] == kClass)
	    nClasses++;

    // Figure out which object this belongs to.  Start off with
    // the lowest inner class, try to resolve the method and if that
    // fails, repeat at a higher class, ending at the top-level class.
    Name actualClass;
    Member *unused;
    while (nClasses > 0) {
	// Hello O(n^2) algorithm.  Hopefully people don't nest their
	// inner classes too deep.
	Name className = _package;
	for (int i = 0;  i < nClasses;  ++i) {
	    if (_types[i] == kClass) {
		if (!className.empty())
		    className += ".";
		className += _names[i];
	    }
	}

	// Lookup the method to see if it really is in this class.
	// If it's inherited, return the inherited method
	if (gSettings.debug)
	    cerr << "--> looking up " << className << "." << name << "...";
	if (_namespace.lookup(className, name, Symbol::kFunction,
			      &actualClass, &unused)) {
	    if (gSettings.debug)
		cerr << endl;
	    return actualClass + string(".") + name;
	}
	// Maybe it's a member and not a method?
	if (_namespace.lookup(className, name, Symbol::kMember,
			      &actualClass, &unused)) {
	    if (gSettings.debug)
		cerr << endl;
	    return actualClass + string(".") + name;
	}

	if (gSettings.debug)
	    cerr << " failed" << endl;

	nClasses--;   // Not in this class;  try one level higher
    }

    // If we get here, the function is probably inherited from
    // a base class we know nothing about.
    return getClass() + string(".") + name;
}
/*
void ScopedNames::addVariable(const Name& name, const Name& type)
{
    if (inMethod())
	_variables[name] = type;
}

bool ScopedNames::lookupVariable(const Name& name, Name *type) const
{
    *type = "";
    if (!inMethod())
	return false;

    std::vector<ScopeType>::const_reverse_iterator rit = _types.rbegin();
    do {
	VariableMap::const_iterator vit = rit->_variables.find(name);
	if (vit != rit->_variables.end()) {
	    *type = vit->second;
	    return true;
	}
	rit++;
    } while (rit != _types.rend() && *rit == kBlock);

    return false;
}
*/
